#define BLOCKSIZE 32

#define SVALUE 0
#define CONSTANT 1
#define SET 2
#define VAR 3
#define ACTION 4
#define FLUENT 5
#define TYPES (FLUENT+1)
#define NVALUE 6
#define OPER 7

typedef enum {optSOLUTIONS,optCONCURRENT,optUNKNOWN,optWARNINGS,
              optINFERENCE,optMUTE} tOPTION;
#define NOPT 6

#define modNONE 0
#define modPERT 1
#define SemError(FORMAT,DATA) {printf("%d: ",yyline); printf(FORMAT,DATA); exit(1);}
#define stEntry(st,pos) (st).Table[(pos)/BLOCKSIZE][(pos)%BLOCKSIZE]
#define ALLOC(type) (type *)malloc(sizeof(type))
#define Tree(n,l,r) btNew(n,sizeof(struct tExpNode),l,r);
#define Leaf(n) btNew(n,sizeof(struct tExpNode),NULL,NULL);
#define Root(t) (*btRoot(t,struct tExpNode))

/* ERROR CODES for pal_code */
#define errNONE 0
#define errBAD_ATOM_ARG 1

typedef unsigned int NUM;
typedef long int SIT;

struct SymTable {
  NUM NSym;
  NUM NBlocks;
  char ***Table,**CurBlock;
};

typedef union {
  NUM Sym;
  int Val;
} ELMVAL;

struct Elem {
  char Type;
  ELMVAL N;
};

struct tFunctHeader {
  NUM RangeSize;
  struct Elem *Range;
  QUEUE Domain;
  NUM Arity,BaseNSym,NTuples;
};

struct tVar {
  NUM RangeSize;
  struct Elem *Range;
};

struct tAtom {
  NUM NSym;
  char *pcktuple;
  struct tFunctHeader *pfh;
  void *data;
};

struct tFunct {
  NUM NSym;
  char Type;
  char Modifier; 	/* modNONE, modPERT */
  SIT Sit; 		/* relative situation 0, -1, -2 ... */
  QUEUE Tuple;
};


struct tExpNode {
  char Type;
  union {
    struct tFunct *Funct;
    char Op;
    NUM Sym;
    int Val;
  } d;
  char isPert;
};

struct tVarLstNode {
  char Type;
  union {
    struct tFunct *Funct;
    NUM Sym;
  } d;
};

typedef struct tQueryNode_ {
  BTREE Expr;
  NUM NVars;
} tQueryNode;

typedef struct tRulInfo_ {
  NUM nRul;
  unsigned line;
} tRulInfo;

struct tGenericFunctions {
  void (*Init)();
  void (*Prepare)();
  void (*InitAtom)(struct tAtom *a,char Type);
  void (*TreatRule)();
  void (*PushSit)();
  void (*PopSit)();
  int (*TreatInit)(NUM atom,NUM value);
  void (*EndInits)();
  int (*TreatObs)(NUM atom,NUM value,char modifier);
  int (*TreatAssign)(NUM atom,NUM value);
  void (*Restart)();
  int (*NewTrans)();
  void (*ShowTrans)();
  int (*GetCurValue)(struct tVarLstNode *pn,struct Elem *pe);
  int (*GetValue)(NUM n,char Type,NUM *v,char *isPert);
  void (*ShowSolution)();
  void (*SolveGroundQuery)();
};

extern struct SymTable SymTables[TYPES];
extern struct SymTable SymData[TYPES];
extern struct SymTable TabAct,TabFlu;
extern struct SymTable tabRulInfo;
extern tRulInfo rulInf;
extern unsigned int yyline;
extern NUM FALSE_NSym,TRUE_NSym,BOOLEAN_NSym,EVENT_NSym;
extern struct Elem *BOOLEAN_RANGE;
extern NUM NSym;
extern NUM DefRangeNSym;
extern char DefType;
extern char AllowNewSVAL;
extern struct tExpNode Node;
extern NUM RuleCount;
extern NUM NGenRules;
extern NUM NRules;
extern QUEUE VarList;
extern struct tVarLstNode *VarVec;
extern BTREE HeadExpr;
extern struct tFunct *HeadFunct;
extern BTREE CondExpr;
extern NUM NVars;
extern struct Elem *tup;
extern NUM NSol;
extern struct tGenericFunctions *GF;
extern char verbose;
extern QUEUE QueryList;
extern tQueryNode *QueryVec;
extern SIT Now;
extern SIT IniSit;
extern SIT ExprSit;
extern FILE *fGroundOut;
extern FILE *fTxtProgram;
extern int UsePipes;
extern int OptVal[NOPT];
extern char yytext[];
extern unsigned pal_code;
extern char pal_errmsg[80];
extern unsigned ShiftSit; /* used for ExprCopy */

void stInit(struct SymTable *st);
NUM stAddSym(struct SymTable *st,char *s);
void stFree(struct SymTable *st);
int GetNSym(char *s,int type,NUM *NSym);
char *strCopy(char *s);
char *strAppend(char *s,char *t);
void GenAtoms(struct tFunctHeader *pf,NUM FunctCode,char Type);
void ComputeIntOper(char op,struct tExpNode *x,struct tExpNode *y);
void DecideOp(BTREE *e,BTREE *e1,BTREE *e2,char op);
void DecideFunct(char *name,char Modifier,SIT Sit,QUEUE params,BTREE *pexp);
int FreeFunct(struct tFunct *pf);
int eqVarLstElm(QUEUE q,void *p2);
void Eval(BTREE e,struct tExpNode *result);
NUM GetAtomNum(struct tFunct *pf);
int GetValueNum(struct tFunct *pf,struct Elem *e,NUM *Num);
int IsFunct(QUEUE q,void *d);
void GenRules();
void RetrieveHead();
int TreatAssign();
int TreatInit();
int Ground (int (*whattodo) ());
void ConcTryState(SIT s);
void NoConcTryState(SIT s);
void SolveGroundQuery();
void SolveQuery();
void EvalAndPrintExpression(BTREE Expr);
void SplitOption(char *arg, char **option, char *type, void **ptrVal);
void SetOpt(char *s,char type,void *ptrVal);
struct tVarLstNode *GetVarLstElm(NUM n);
int TupleCopy(QUEUE q,void *d);
struct tFunct *FunctCopy(struct tFunct *pf);
BTREE ExprCopy(BTREE e);
struct tGenericFunctions *GetGF(int val);
void WriteGroundAtoms(struct tFunctHeader *pf,char Type);
void WriteGroundRule(unsigned Num);
void WriteGroundQuery();


char *Elem2str(struct Elem *e);
void showElem(struct Elem *e);
char *Value2str(struct tExpNode *e);
void showVALUE(struct tExpNode *e);
char *Atom2str(struct tAtom *a,char Type);
char *AtomNum2str(NUM i,char Type);
void showAtom(NUM i,char Type);
char *AtomVal2str (NUM i,NUM v,char Type);
void showAtomVal(NUM i,NUM v,char Type);
void showAtoms(struct tFunctHeader *pf,char Type);
void printbanner();
void printhelp();
void showFunct(struct tFunct *pf);
int showVarLstElm(QUEUE q,void *d);
void showRule(unsigned Num);
unsigned GetRulNum(NUM r);
int InsertNewVar(QUEUE *pVList,char Type,NUM NSym,struct tFunct *pf);
int showTree(BTREE t);
