#include "all.h"
#include <sys/time.h>
#ifdef RUSAGE
#include <sys/resource.h>
#endif
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h> 
#include <sys/stat.h> 
#include <sys/file.h>

extern FILE *yyout;
int main (int argc,char *argv[])
{
  FILE *File;
  
  extern char *optarg;
  extern int optind;
  int c;
  char k;
  int argpos=1;
  int fifoin,fifoout;
  FILE *fin,*fout;
  
  #ifdef RUSAGE
  struct rusage r;
  #else
  struct timeval tv1,tv2;
  struct timezone tz = {0,DST_NONE};
  double secs;
    
  gettimeofday(&tv1,&tz);
  #endif
  
  /* Processing options */
  
  while ((c = getopt(argc, argv, "g:hmo:p:F")) != EOF) {
    switch (c) {
      case 'o': 
        {
	  char *optName,type;
	  void *optVal;
          SplitOption(optarg,&optName,&type,&optVal);
	  SetOpt(optName,type,optVal);
	  free(optName);
	  free(optVal);	 
	  argpos+=2;
	}  
        break;
      case 'm':
        OptVal[optMUTE]=1;
	argpos++;
        break;
      case 'g':
        {
	  if ((fGroundOut=fopen(optarg,"w"))==NULL) {
	    printf("Cannot open file %s\n",argv[argpos]);
	    exit(0);
	  }
	  argpos+=2;
	}
	break;
      case 'p':
        {
	  if ((fTxtProgram=fopen(optarg,"w"))==NULL) {
	    printf("Cannot open file %s\n",argv[argpos]);
	    exit(0);
	  }
	  argpos+=2;
	}
	break;
      case 'F':
        {
	  UsePipes=1;
	  if ((fifoin = open("pal.in.fifo", O_RDONLY)) < 0) {
	      perror("open pal.in.fifo");
	      exit(1);
	  }
	  if ((fifoout = open("pal.out.fifo", O_WRONLY)) < 0) {
	      perror("open pal.out.fifo");
	      exit(1);
	  }
	  argpos++;
	}
	break;
      case 'h':
      case '?': printhelp();   
    }
  }
  
  if (!OptVal[optMUTE])
    printbanner();
  if (argpos<argc) {
    if ((File=fopen(argv[argpos],"r"))==NULL) {
      printf("Cannot open file %s\n",argv[argpos]);
      exit(0);
    }
    yyrestart(File);
    if (!OptVal[optMUTE])
      printf("loading file %s...\n",argv[argpos]);
  }  
  
  PALInit();

  yyparse();
  
  if (!OptVal[optMUTE]) {
    printf("%d actions, %d fluents, ",
           TabAct.NSym,TabFlu.NSym);
    printf("%d basic rules, ",NRules);

    #ifdef RUSAGE
    getrusage (RUSAGE_SELF, &r);
    printf("%ld.%03ld seconds.\n",r.ru_utime.tv_sec,r.ru_utime.tv_usec/1000);
    #else
    gettimeofday(&tv2,&tz);  
    if (tv2.tv_sec==tv1.tv_sec)
      secs=(tv2.tv_usec-tv1.tv_usec)*1e-6;    
    else
      secs=(tv2.tv_sec-tv1.tv_sec-1)+
           (tv2.tv_usec+(tv1.tv_sec)?1000000L-tv1.tv_usec:0)*1e-6;
    printf("%2.3f seconds.\n",secs);
    #endif
  }
  
  if (argc>1) {
    yyrestart(stdin);
    if (!UsePipes)
      yyparse();
    else
      while (1) {
        read(fifoin,&k,1);
	if (k==0) break;
	/* redirect input and output */
	if ((fin = fopen("pal.in", "r")) < 0) {
	    perror("fopen pal.in");
	    exit(1);
	}
        yyrestart(fin);

	if ((fout = fopen("pal.out", "w")) < 0) {
	    perror("fopen pal.out");
	    exit(1);
	}
	yyout=fout;
	
	/* Parse sentence */    
	yyparse();
	fclose(fin);
	fclose(fout);
	
	/* Warn that parsing has finished */
	write(fifoout,&k,1);
      }
  }
    
  PALDone();
    
  if (fGroundOut!=NULL) fclose(fGroundOut);
  if (fTxtProgram!=NULL) fclose(fTxtProgram);
  
}

