#include "all.h"

/*#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "queue.h"
#include "btree.h"
#include "binvec.h"
#include "icode.h"
#include "palmc.h"
#include "mcs.h"
#include "sstack.h"
*/

struct tGenericFunctions MCGF = 
  {MCInit,
   MCPrepare,
   MCInitAtom,
   MCTreatRule,
   MCPushSit,
   MCPopSit,
   MCTreatInit,
   MCEndInits,
   MCTreatObs,
   MCTreatAssign,
   MCRestart,
   MCNewTrans,
   MCShowTrans,
   MCGetCurValue,
   MCShowSolution,
   SolveGroundQuery};

/* Local variables */

static char NoInputs=1;
static char Initially=1;


tGMDATA FalseData;
#define GMINITIAL(TYPE) ((TYPE==ACTION)?'a':'f')
#define DATA(CODE,TYPE) ((TYPE==ACTION)?((struct tAtom*)stEntry(TabAct,CODE))->data:((struct tAtom*)stEntry(TabFlu,CODE))->data)




void MCInit() {
  char s[]="x_false";
  CreateMG(s,FLUENT);
  FalseData.RulCnt=0;
}


void MCInitAtom(struct tAtom *a, char Type)
{
  char s[100];
  
  if (Type==ACTION)
    sprintf(s,"%c_%d",GMINITIAL(Type),TabAct.NSym-1);
  else
    sprintf(s,"%c_%d",GMINITIAL(Type),TabFlu.NSym-1);
  a->data=(void *)malloc(sizeof(tGMDATA));
  ((tGMDATA*)a->data)->RulCnt=0;
  CreateMG(s,Type);
  #ifdef DEBUG
  printf("<MEDTOOL> create MG %s\n",s);
  #endif
}

void MCTreatRule()
{
  unsigned i;
  struct Elem e;
  struct tExpNode Node;
  NUM n,v;
  
  char *KM,HeadGM[100],s[100];

  /* Create new Rule structure */
  
  /* Treat head */    
  
  if (HeadFunct==NULL)
    /* this is a rule whose head is 'false' */
    {
      sprintf(HeadGM,"x_false_%d",FalseData.RulCnt++);
      KM=strCopy("0 if 1");
    }
  else {
    if (!(HeadFunct->Type==FLUENT 
          && HeadFunct->Modifier==modNONE
	  && HeadFunct->Sit==0)) {
      printf("%d: rule head must be a fluent.\n",yyline); 
      exit(1);
    }  
    Eval(HeadExpr,&Node);
    e.Type=Node.Type;
    switch (e.Type) {
      case SVALUE : e.N.Sym=Node.d.Sym; break;
      case NVALUE : e.N.Val=Node.d.Val; break;
    }
    if (GetValueNum(HeadFunct,&e,&v)) {
      n=GetAtomNum(HeadFunct);
      sprintf(HeadGM,"%c_%d_%d",
      		GMINITIAL(HeadFunct->Type),
      		n,((tGMDATA *)DATA(n,HeadFunct->Type)) -> RulCnt ++		
		);
		
      sprintf(s,"%d if 1",v);
      KM=strCopy(s);

    }
    else {
      if (OptVal[optWARNINGS])
      {
	printf("Warning: head `");
	showAtom(GetAtomNum(HeadFunct),HeadFunct->Type);
	printf(":=");
	showElem(&e);
	printf("' out of range in rule %d. Converted into `false'.\n",
  	       RuleCount,yyline);       
      }

      /* the rule head must be considered 'false' */
      sprintf(HeadGM,"x_false_%d",FalseData.RulCnt++);    
      KM=strCopy("0 if 1");
    }
  }

  /* Treat body */

  for (i=0;i<NVars;i++)
    if (VarVec[i].Type!=VAR) {
      n=GetAtomNum(VarVec[i].d.Funct);
      GetValueNum(VarVec[i].d.Funct,&tup[i],&v);
      if (VarVec[i].d.Funct->Sit==-1)
        sprintf(s," and previous(%c_%d)=%d",GMINITIAL(VarVec[i].Type),n,v);
      else
        sprintf(s," and %c_%d=%d",GMINITIAL(VarVec[i].Type),n,v);
      KM=strAppend(KM,s);
    }

  /* Store rule */
  KM=strAppend(KM,";");
  CreateMG(HeadGM,FLUENT);
  AddKM(HeadGM,KM);
  #ifdef DEBUG
  printf("<MEDTOOL> RULE %s:=%s\n",HeadGM,KM);
  #endif
  free(KM);
}

void MCPrepare()
{
  NUM i,j;
  char *KM,s[100];
  static char First=1;
  
  if (First) First=0; else return;

  for (i=0;i<TabFlu.NSym;i++) {
    KM=strCopy("");
    for (j=0;j<((tGMDATA *)DATA(i,FLUENT))->RulCnt; j++) {
      sprintf(s,"f_%d_%d if timeof(now) and pertinent(f_%d_%d);\n",i,j,i,j);
      KM=strAppend(KM,s);
    }
    sprintf(s,"f_%d",i);
    AddKM(s,KM);
    #ifdef DEBUG
    printf("<MEDTOOL> RULE %s:=%s\n",s,KM);
    #endif
    free(KM);  
  }
  
  SaveES();
}


void MCRestart()
{
  /*  This "forgets" current state and returns data to creation time */
  Now=0;
  ResetES();
  #ifdef DEBUG
  printf("<MEDTOOL> Reset Expert System\n");
  #endif
  printf("\nRestart\n");
}


void MCPushSit()
{
  Now++;
}


void MCPopSit()
{
  Now--;
}


int MCTreatInit(NUM atom,NUM value)
{
  char s[100];

  /* if we had given another DIFFERENT value
  if ( ) {
  
    if (verbose) {
      printf("Inconsistence (initial values for fluent `");
      showAtom(atom,FLUENT);
      printf("')\n"); 
    }
    return 0;
  }
  */
  
  
  /* assing initial value */  
  sprintf(s,"f_%d",atom);  
  /*AddMGInitValue(s,(double)value);*/
  PutMGValue(s,(double)value);
  #ifdef DEBUG
  printf("<MEDTOOL> INIT VAL %s=%f\n",s,(double)value);
  #endif
  return 1;
}

void MCEndInits()
{
  FireTrans();
  #ifdef DEBUG
  printf("<MEDTOOL> Fired transition for initial values\n");
  #endif
}


int MCTreatObs(NUM atom,NUM value,char modifier)
{
  /* store the observation */
  /* modifier can be modNONE, i.e., the observation is about the value
    or modPERT, i.e. it is about pertinence */

  /* After execution MCPushSit, all the observations must be checked */

  return 1;  
}


int MCTreatAssign(NUM atom,NUM value)
{
  char s[100];

  sprintf(s,"a_%d",atom);
    
  NoInputs=0;
/*
  if (AlreadyInputMG(s)) {
    if (verbose) {
      printf("Inconsistence (values for action `");
      showAtom(atom,ACTION);
      printf("')\n"); 
    }
    return 0;
  }
  */
  /* do the input */
  PutMGValue(s,(double)value);  
  #ifdef DEBUG
  printf("<MEDTOOL> %s:=%u\n",s,value);
  #endif
  return 1;
}


void ShowGM(char *name) {
  NUM n;
  double d;
  char type;
  char *s;
  
  /*if (strchr(&name[2],'_')!=NULL) return;*/

  if (name[0]=='f' && GetMGUnknownValue(name)) {
    n=atoi(&name[2]);
    printf("%s:=?\n",s=AtomNum2str(n,FLUENT));
    free(s);
  }
  else
    if (GetMGPertinentValue(name)) {
      n=atoi(&name[2]);
      type=(name[0]=='a')?ACTION:FLUENT;
      printf("%s:=",s=AtomNum2str(n,type));
      free(s);
/*      
      if (GetMGUnknownValue(name))
        printf("?\n");
      else {
	GetMGValue(name,&d);
	printf("%s\n",s=AtomVal2str(n,(NUM)d,type));
	free(s);
      }
*/      
      GetMGValue(name,&d);
      printf("%s\n",s=AtomVal2str(n,(NUM)d,type));
      free(s);
    }
}

void MCShowTrans() {
  NUM i;
  char s[100];
  
  printf("%d)\n",Now);
  /*ShowPertinents(ShowGM);*/
  for(i=0;i<TabAct.NSym;i++) {
    sprintf(s,"a_%d",i);
    if (GetMGPertinentValue(s))    
      ShowGM(s);
  }
  for(i=0;i<TabFlu.NSym;i++) {
    sprintf(s,"f_%d",i);
    if (GetMGPertinentValue(s))    
      ShowGM(s);
  }
    
  
}

int MCNewTrans()
{
  char ok=1;
  /* Fire an evaluation */

  if (NoInputs) PutInNewDataSituation();
  

  FireTrans();
  #ifdef DEBUG
  printf("<MEDTOOL> Fired new transition\n");
  #endif
  NoInputs=1;

  /* ok should return 0 if inconsistence reached or observations not
     satisfied
  */
  return ok;
}


int MCGetCurValue(struct tVarLstNode *pn,struct Elem *pe)
{
  /* return current value of a fluent or action */

  struct tFunctHeader *pf;
  NUM n;
  double d;
  char s[100];


  n=GetAtomNum(pn->d.Funct);
  sprintf(s,"%c_%d",GMINITIAL(pn->Type),n);
  
  #ifdef DEBUG
  printf("<MEDTOOL> Asking value of GM %s\n",s);
  #endif
  
  switch(pn->d.Funct->Modifier) {
    case modNONE:
      if (pn->d.Funct->Sit==0) {
	if (pn->Type==ACTION && !GetMGPertinentValue(s)) return 0;
	if (GetMGUnknownValue(s)) return 0;
	GetMGValue(s,&d);
	if (pn->Type==ACTION)
	  pf=((struct tAtom *)stEntry(TabAct,n))->pfh;
	else
	  pf=((struct tAtom *)stEntry(TabFlu,n))->pfh;
	memcpy(pe,&pf->Range[(NUM)d],sizeof(struct Elem));
      }
      else {
	/* it is the previous value of the action */
      }
      break;
    case modPERT:
      pe->Type=SVALUE;
      if (pn->d.Funct->Sit==0) {
        if (GetMGPertinentValue(s))
	  pe->N.Sym=TRUE_NSym;
	else
	  pe->N.Sym=FALSE_NSym;
      }
      else {
	/* if (action was prev pertinent) */
      }
      break;
  }
  
  return 1;
}


void MCShowSolution()
{
  NUM i;
  SIT s;
  char first=1;
  QUEUE *FluData=NULL,*ActData=NULL;
  void *p;
  
  NSol++;
  if(ExprSit>1)
    printf("\nSolution %d:\n",NSol);
  for(i=0;i<NVars;i++)
    if (VarVec[i].Type==VAR)
    {
      if (first) first=0; else printf(",");
      printf("%s",stEntry(SymTables[VAR],VarVec[i].d.Sym));
      printf("=");
      showElem(&tup[i]);
    }
  if (!first) printf("\n");
  
  if (ExprSit<=1) return; /* no temporal query */
  
  /* for (s=IniSit+1;s<IniSit+ExprSit;s++) */
  /* show executed actions at situation s */
}
