/* Scanner for pal */

%{
#include <math.h>
#include <string.h>
#include "queue.h"
#include "btree.h"
#include "binvec.h"
#include "icode.h"
#include "sets.h"
#include "y.tab.h"	/* Tokens definitions generated by yacc (option -d)*/

extern int yyerror(char *);
%}
%option noyywrap
%array

DIGIT    [0-9]
ID       [a-z][a-zA-Z0-9_]*
VID      [A-Z][a-zA-Z0-9_]*
STRING   "'"[^']*"'"

%%

{DIGIT}+	{ yylval.num = atoi(yytext); return NUMBER; }

x		{ yylval.txt = (char *) malloc(strlen(yytext) + 1);
		  strcpy(yylval.txt,yytext);
		  return EKS; }
constants	{ return CONSTANTS; }
sets		{ return SETS; }
actions		{ return ACTIONS; }
fluents		{ return FLUENTS; }
vars		{ return VARS; }
rules		{ return RULES; }
initially	{ return INITIALLY; }
init		{ return INITIALLY; }
query		{ return QUERY; }
do		{ return DO; }
if		{ return IF; }
and		{ return AND; }
or		{ return OR; }
not		{ return NOT; }
pert		{ return PERT; }
pertinent	{ return PERT; }
prev		{ return PREV; }
previous	{ return PREV; }
true		{ return TRUE; }
false		{ return FALSE; }
options		{ return OPTIONS; }
print		{ return PRINT;}

{ID}		{ yylval.txt = (char *) malloc(strlen(yytext) + 1);
		  strcpy(yylval.txt,yytext);
		  return ID;}
{VID}		{ yylval.txt = (char *) malloc(strlen(yytext) + 1);
		  strcpy(yylval.txt,yytext);
		  return VID;}
{STRING}	{ 
                  int len=strlen(yytext)-2;
                  yylval.txt = (char *) malloc(len+1);
		  yylval.txt[len]='\0';
		  if (len)		  		  
  		    strncpy(yylval.txt,yytext+1,len);
		  return STRING;}
"..."		{ return LATER; }		  
":="		{ return ASSIGN; }
"->"		{ return ARROW;}
">="		{ return GEQ; }	
"<="		{ return LEQ; }	
"!="		{ return NEQ; }	
"&&"		{ return SAND; }
"||"		{ return SOR; }
"/*"([^*]|"*"[^/])*"*/"	{
		 char *s;
		 for (s=yytext;(s=strchr(s,'\n'))!=NULL; s++,yyline++);
		}
"#"[^\n]*	{} /* eat up one-line comments */

[ \t\r]+	/* eat up whitespace */
\n		{yyline++;}
.           	{return yytext[0];}

%%
