%{
/*
#include <math.h>	
#include <search.h>
#include <stdio.h>
#include <stdlib.h>
#include "queue.h"
#include "btree.h"
#include "binvec.h"
#include "icode.h"
#include "sets.h"
#include "wfs.h"
*/
#include "all.h"
#include <unistd.h>

extern int yylex(void);
extern int yyrestart(FILE *);
int yyerror (char *s)  /* Called by yyparse on error */
{
  printf ("%d: %s after `%s'\n", yyline, s, yytext);
  return 1;
}

%}

%union {
int	num;		/* For returning numbers	*/
char  	*txt;		/* For returning text		*/
struct Elem elem;
struct tFunctHeader ftype;
QUEUE set;
QUEUE list;
BTREE expr;
tQueryNode *qnode;
}

%token  NUMBER     /* integer   */
%token  ID         /* identifier */
%token  VID        /* variable identifier */
%token  EKS        /* identifier 'x'*/
%token  STRING     /* string literal */
%token  SETS
%token  CONSTANTS
%token  ACTIONS
%token  FLUENTS
%token  VARS
%token  RULES
%token  INITIALLY
%token  QUERY
%token  LATER
%token  DO
%token  ASSIGN
%token  ARROW
%token  TRUE
%token  FALSE
%token  OPTIONS
%token  PRINT
%token  SAND
%token  SOR

%type <set> setexpr
%type <set> rangeforzeroary
%type <set> set
%type <set> values
%type <elem> value
%type <elem> constant
%type <ftype> fheader
%type <list> tuple
%type <list> vidlist
%type <list> idlist
%type <txt> id
%type <txt> vid
%type <expr> exp
%type <expr> funct
%type <list> params
%type <qnode> querynode

%left IF
%left OR SOR
%left AND SAND
%left NOT
%left PERT PREV
%right '=' LEQ GEQ NEQ '>' '<'
%left MAX MIN
%left '-' '+'
%left '*' '/' '%'
%left NEG     /* Negation--unary minus */

/* Grammar follows */

%%

program : declarations 
         { static char first=1;
	   if (first) {
	     GF->Prepare(); 
	     first=0;
	   }
	 } 
	 sentences ;

declarations : 
	| declarations declaration ;
	
sentences : 
	| sentences sentence ;	

declaration: CONSTANTS {AllowNewSVAL=1;} constdefs {AllowNewSVAL=0;}
	| SETS {AllowNewSVAL=1;} setdefs {AllowNewSVAL=0;}
	| ACTIONS {DefRangeNSym=EVENT_NSym;
		   AllowNewSVAL=1; 
		   DefType=ACTION;} fsymbols
		  {AllowNewSVAL=0;}
	| FLUENTS {DefRangeNSym=BOOLEAN_NSym;
		   AllowNewSVAL=1; 
		   DefType=FLUENT;} fsymbols
		  {AllowNewSVAL=0;}
	| VARS vars
	| RULES rules
	| OPTIONS options ';'
	;
	
sentence : INITIALLY 
		{ GF->Restart();
		  if (fGroundOut) fprintf(fGroundOut,"\n\ninitially\n");
		} 
	   initials ';' 
	   	{ GF->EndInits(); 
		  if (fGroundOut) fprintf(fGroundOut,";\n");
		}
	| DO '{' execs '}'
	| QUERY queries
	;

options : option
	| options ',' option

option  : id '=' NUMBER		{int i;
				 i=*(int *)&yylval.num;
                                 SetOpt($1,NVALUE,&i);
				}
	| id '=' id		{SetOpt($1,SVALUE,$3);}
	| id '=' TRUE		{int i=1; SetOpt($1,NVALUE,&i);}
	| id '=' FALSE		{int i=0; SetOpt($1,NVALUE,&i);}
	| id			{int i=1; SetOpt($1,NVALUE,&i);}
	| NOT id		{int i=0; SetOpt($2,NVALUE,&i);}
	;
	

constdefs : constdef ';'
	| constdefs constdef ';'
	;

constdef : id '=' exp		{
				 struct Elem *pe=ALLOC(struct Elem);
				 if (GetNSym($1,CONSTANT,&NSym)>=0)
				   SemError("`%s' already defined.\n",$1);

				 switch (Root($3).Type) {
				   case NVALUE:
				     pe->N.Val=Root($3).d.Val; break;	
				   case SVALUE:
				     pe->N.Sym=Root($3).d.Sym; break;
				   default:
				   SemError("expression for `%s' is not constant.\n",$1);
				 }
				 pe->Type=Root($3).Type;
				 #ifdef DEBUG
				 printf("%s=",$1);
				 showElem(pe);
				 printf("\n");
				 #endif
				 stAddSym(&SymData[CONSTANT],(char *)pe);
				 free($1);
				 btFree(&$3);
				}
	;

setdefs : setdef ';'
	| setdefs setdef ';'
	;

setdef	: id '=' setexpr 
				{
				 #ifdef DEBUG
				 printf("%s=",$1); setShow($3); printf("\n");
				 #endif
				 if (GetNSym($1,SET,&NSym)>=0)
				   SemError("`%s' already defined.\n",$1);
				 stAddSym(&SymData[SET],(char *)$3);
				 free($1);
				}
	;

setexpr	: set			{$$=$1;}
	| '(' setexpr ')'	{$$=$2;}
	| setexpr '*' setexpr	{$$=setIntersec($1,$3);
				 setFree(&$1); setFree(&$3);
				}
	| setexpr '+' setexpr	{$$=setUnion($1,$3);
			         setFree(&$1); setFree(&$3);
				}

	| setexpr '-' setexpr	{$$=setDiff($1,$3);
				 setFree(&$1); setFree(&$3);
				}

	;

set	: '{' values '}'	{$$=$2;}
	| '{' '}'		{setEmpty(&$$);}
	| id			{if (GetNSym($1,SET,&NSym)!=SET) 
				   SemError("`%s' is not defined or is not a set\n",$1);
				 $$=setCopy(
				 (QUEUE)stEntry(SymData[SET],NSym));
				 free($1);
				}
	| '[' exp ',' exp ']'
				{struct Elem e;
				 if (Root($2).Type!=NVALUE||
				     Root($4).Type!=NVALUE)
				  {printf("Semantic error: non numeric interval\n");
				   exit(1);				   
				  }				  
				 setEmpty(&$$);
				 e.Type=NVALUE;
				 e.N.Val=Root($2).d.Val;
				 while (e.N.Val <= Root($4).d.Val) {
				   setAddElem(&$$,&e);
				   e.N.Val++;
				 }
				 btFree(&$2);
				 btFree(&$4);
				}

	;
	
values 	: value			{setEmpty(&$$);
				 setAddElem(&$$,&$1);
				}
	| values ',' value	{setAddElem(&$1,&$3);
				 $$=$1;
				}
	;

value	: exp			{switch($$.Type=Root($1).Type) {
				 case NVALUE:
				   $$.N.Val=Root($1).d.Val; break;
				 case SVALUE:
				   $$.N.Sym=Root($1).d.Sym; break;
				 default:
				   SemError("set value must be a constant expression%s\n","");
				 }
				 btFree(&$1);
				}

constant : NUMBER		{$$.Type=NVALUE; 
				 $$.N.Sym=*(NUM *)&yylval.num;
				}
	| TRUE			{$$.Type=SVALUE;
				 $$.N.Sym=TRUE_NSym;
				}
	| FALSE			{$$.Type=SVALUE;
				 $$.N.Sym=FALSE_NSym;
				}
	;
	

fsymbols : fsymbol ';'
	| fsymbols fsymbol ';'
	;
	
fsymbol : idlist fheader	{char *s;
				 struct tFunctHeader *pfh;
				 #ifdef DEBUG
				 NUM i;
				 #endif 
				 while (qIsNotEmpty($1)) {
				   s=*qFirst($1,char *);
				   
				   qDeleteFirst(&$1);

				   if (GetNSym(s,DefType,&NSym)>=0)
				     SemError("`%s' already defined.\n",s)
				   pfh=(struct tFunctHeader *)
				       malloc(sizeof(struct tFunctHeader));
				   memcpy(pfh,&$2,sizeof(struct tFunctHeader));
				   pfh->Arity=qCount(pfh->Domain);
				   stAddSym(&SymData[DefType],(char *)pfh);
				   GenAtoms(pfh,
				        SymData[DefType].NSym-1,DefType);
				   #ifdef DEBUG
				   printf("action/fluent %s\n",s);
				   printf("   domain ");ShowDomain(pfh->Domain);
				   printf("\n   range {");
				   for (i=0;i<=pfh->RangeSize-1;i++) {
				     showElem(&pfh->Range[i]);
				     printf(" ");
				   }
				   printf("}\n");
				   showAtoms(pfh,DefType);
				   printf("\n");
				   #endif
				   if (fGroundOut)
				     WriteGroundAtoms(pfh,DefType);
				   free(s);
				 }
				}
	;

fheader :			{$$.Domain=qEMPTY;
				 $$.Range=(struct Elem *)
				   Queue2Array(
				     (QUEUE)stEntry(SymData[SET],DefRangeNSym),
				      sizeof(struct Elem),
				      &$$.RangeSize);
				}
	| ':' rangeforzeroary	{$$.Domain=qEMPTY;
				 $$.Range=(struct Elem *)
				   Queue2Array($2,sizeof(struct Elem),
				               &$$.RangeSize);
				 qFree(&$2);
				}
	| ':' tuple ARROW setexpr
				{$$.Domain=$2; 
				 $$.Range=(struct Elem *)
				   Queue2Array($4,sizeof(struct Elem),
				               &$$.RangeSize);
				 qFree(&$4);
				}
	;

rangeforzeroary : 
          ARROW setexpr		{$$=$2;}
	| setexpr 		{$$=$1;}
	;
	
tuple	: setexpr		{$$=qEMPTY;
				 qInsertLast(&$$,&$1,sizeof(QUEUE));
				}
	| tuple EKS setexpr 	{qInsertLast(&$1,&$3,sizeof(QUEUE));
				 $$=$1;
				}
	;
		

vars	: var ';'
	| vars var ';'
	;

var	: vidlist ':' setexpr	{char *s;
				 struct tVar *pv;
				 
				 while(qIsNotEmpty($1))				 				 {
				   s=*qFirst($1,char *);
				   qDeleteFirst(&$1);
				   if (GetNSym(s,VAR,&NSym)>=0)
				     SemError("`%s' already defined.\n",s);
				   pv=(struct tVar*)malloc(sizeof(struct tVar));					   pv->Range=(struct Elem *)
				     Queue2Array($3,
				            sizeof(struct Elem),&pv->RangeSize);
				   #ifdef DEBUG
				   printf("var %s\n",s);
				   printf("   range ");setShow($3); 
				   printf("\n");
				   #endif
				   stAddSym(&SymData[VAR],(char *)pv);
				   free(s);				   
				 }
				 qFree(&$3);
				}
	;
	
rules	: rule0 ';'
	| rules rule0 ';'
	;
	
rule0 	: 
		{NVars=0; RuleCount++;
		 rulInf.line=yyline;
		 #ifdef DEBUG
		 printf("\nRULE %d\n",RuleCount);
		 #endif
		 NGenRules=0;
		}
	  rule
		{
		  tRulInfo *ri;
		  /* qBrowse(VarList,showVarLstElm,NULL);printf("\n");*/
		  GenRules();	
		  if (!NGenRules && OptVal[optWARNINGS])
		    printf("Warning: rule %d is never applicable.\n",RuleCount);
		  rulInf.nRul=NGenRules;
		  ri=(tRulInfo*)malloc(sizeof(tRulInfo));
		  memcpy(ri,&rulInf,sizeof(tRulInfo));		  
		  stAddSym(&tabRulInfo,(char *)ri);		  
		} 
	;	
	
initials : initial
	| initials ',' initial
	;
	
initial : {NVars=0;} rule {if (!Ground(TreatInit)) exit(0);} 
	;	

execs   : exec
	| execs exec	
	;

exec	: {GF->PushSit();} exec2 ';' {
                                      if (!GF->NewTrans()) GF->PopSit();
				      else GF->ShowTrans();
				     }
	| PRINT '(' printexprs ')' ';' {printf("\n");}
	| PRINT ';' {printf("\n");}
	;

printexprs: printexpr
	| printexprs ',' printexpr ;

printexpr: {NVars=0;} 
  	    exp 
           {EvalAndPrintExpression($2);btFree(&$2);} 
	 | STRING 
	   { printf("%s",yylval.txt);
	     free(yylval.txt);
           }
	 ;

exec2	: 
	| actions
	;
		
actions	: action
	| actions ',' action
	;	

action	: {NVars=0;} rule {if (!Ground(TreatAssign)) exit(0);}
	;
	
rule	: assign	 {CondExpr=NULL;}
	| assign IF exp  {CondExpr=$3;}
	;

assign	: funct {RetrieveHead();} ASSIGN exp 
                        {HeadExpr=$4; 
			 btFree(&$1);
			}
	| FALSE		{HeadExpr=NULL; HeadFunct=NULL;}
	| funct		{RetrieveHead();
			 Node.Type=SVALUE;
			 Node.d.Sym=TRUE_NSym;
			 HeadExpr=Leaf(&Node);
			 btFree(&$1);
			}
	| NOT funct	{RetrieveHead();
			 Node.Type=SVALUE;
			 Node.d.Sym=FALSE_NSym;
			 HeadExpr=Leaf(&Node);
			 btFree(&$2);
			}
	;

queries	: query
	| queries query
	;

query	: {NVars=0;} querystates '?'
			{SolveQuery();}

querystates : queryexp
	| querystates ';' queryexp
	;

queryexp : querynode
				{
				 qInsertLast(&QueryList,$1,sizeof(tQueryNode));
				 free($1);
				 ExprSit++; 
				}

	| querynode LATER '{' exp
				{
				 SIT i;
				 
				 if (Root($4).Type!=NVALUE ||
				     Root($4).d.Val<0) {
				   printf("%d: ", yyline);
				   printf("invalid number for ");
				   printf("'...' operator\n");
				   exit(0);
				 }

				 /*create many copies of expression*/
				 for (i=0;i<Root($4).d.Val;i++) {
				   qInsertLast(&QueryList,$1,
				               sizeof(tQueryNode));
				   ExprSit++; 	
				   ShiftSit=1;
				   $1->Expr=ExprCopy($1->Expr);
				   ShiftSit=0;
				   $1->NVars=NVars;
				 }
				 HeadExpr=$1->Expr;
				 free($1);
				}
	  '}' querynode
				{
				 BTREE e;
				 DecideOp(&e,&$7->Expr,&HeadExpr,'^');
				 $7->Expr=e;
				 qInsertLast(&QueryList,$7,sizeof(tQueryNode));
				 ExprSit++;
				 free($7);
				}				
	;


querynode :			{Node.Type=SVALUE;
				 Node.d.Sym=TRUE_NSym;
				 $$=(tQueryNode*)malloc(sizeof(tQueryNode));
				 $$->Expr=Leaf(&Node);
				 $$->NVars=NVars;
				}
	| exp			{
				 $$=(tQueryNode*)malloc(sizeof(tQueryNode));
				 $$->Expr=$1;
				 $$->NVars=NVars;
				}
	;

exp 	: funct		{$$=$1;}
	| vid		{if (GetNSym($1,VAR,&NSym)!=VAR)
			   SemError("undefined variable `%s'.\n",$1);
			 Node.Type=VAR;
		         Node.d.Sym=
		            InsertNewVar(&VarList,VAR,NSym,NULL);
			 $$=Leaf(&Node);
			 
			 free($1);
			}
	| constant	{
			 if ((Node.Type=$1.Type)==NVALUE)
			   Node.d.Val=$1.N.Val;
			 else
			   Node.d.Sym=$1.N.Sym;
			 $$=Leaf(&Node);
			}
	| '(' exp ')'	{$$=$2;}

	| exp AND exp   {DecideOp(&$$,&$1,&$3,'^');}
	| exp SAND exp   {DecideOp(&$$,&$1,&$3,'&');}
	| exp OR exp    {DecideOp(&$$,&$1,&$3,'v');}
	| exp SOR exp    {DecideOp(&$$,&$1,&$3,'|');}
	| NOT exp       {DecideOp(&$$,NULL,&$2,'');}

	| exp '>' exp	{DecideOp(&$$,&$1,&$3,'>');}
	| exp '<' exp	{DecideOp(&$$,&$1,&$3,'<');}
	| exp '=' exp	{DecideOp(&$$,&$1,&$3,'=');}
	| exp GEQ exp	{DecideOp(&$$,&$1,&$3,'g');}
	| exp LEQ exp	{DecideOp(&$$,&$1,&$3,'l');}
	| exp NEQ exp	{DecideOp(&$$,&$1,&$3,'n');}	

	| exp '+' exp	{DecideOp(&$$,&$1,&$3,'+');}
	| exp '-' exp	{DecideOp(&$$,&$1,&$3,'-');}
	| exp '*' exp	{DecideOp(&$$,&$1,&$3,'*');}
	| exp '/' exp	{DecideOp(&$$,&$1,&$3,'/');}
	| exp '%' exp	{DecideOp(&$$,&$1,&$3,'%');}
	| '|' exp '|'	{DecideOp(&$$,NULL,&$2,'a');}
        | '-' exp  %prec NEG
			{DecideOp(&$$,NULL,&$2,'m');}
	;

funct 	: id		{DecideFunct($1,modNONE,ExprSit,NULL,&$$);}
	| id '(' params ')'		
			{DecideFunct($1,modNONE,ExprSit,$3,&$$);}
	| PREV '(' id ')'
			{DecideFunct($3,modNONE,ExprSit-1,NULL,&$$);}
	| PREV '(' id '(' params ')' ')'
			{DecideFunct($3,modNONE,ExprSit-1,$5,&$$);}
	| PREV '(' PERT '(' id ')' ')'
			{DecideFunct($5,modPERT,ExprSit-1,NULL,&$$);}
	| PREV '(' PERT '(' id '(' params ')' ')' ')'
			{DecideFunct($5,modPERT,ExprSit-1,$7,&$$);}
	| PERT '(' id ')'		
			{DecideFunct($3,modPERT,ExprSit,NULL,&$$);}
	| PERT '(' id '(' params ')' ')'
			{DecideFunct($3,modPERT,ExprSit,$5,&$$);}
	;

params  : exp		 {
			   $$=qEMPTY;
			   qInsertLast(&$$,&$1,sizeof(BTREE));
			 }
	| params ',' exp {
			   $$=$1;
			   qInsertLast(&$$,&$3,sizeof(BTREE));
			 }
	;

idlist  : id			{$$=qEMPTY;
				 qInsertLast(&$$,&$1,sizeof(char *));}
	| idlist ',' id	        {qInsertLast(&$1,&$3,sizeof(char *));
				 $$=$1;
				}
	;

id	: ID			{$$=yylval.txt;}
	| EKS			{$$=yylval.txt;}
	;

vidlist : vid			{$$=qEMPTY;
				 qInsertLast(&$$,&$1,sizeof(char *));}
	| vidlist ',' vid	{qInsertLast(&$1,&$3,sizeof(char *));
				 $$=$1;
				}
	;

vid	: VID			{$$=yylval.txt;}
	;

/* End of grammar */
%%

