#include <iostream.h>
#include "smodels.h"
#include "api.h"
#include "atomrule.h"
#include "all.h"

/*#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "queue.h"
#include "btree.h"
#include "binvec.h"
#include "icode.h"
#include "sms.h"
#include "sstack.h"
*/

typedef struct {
  char Modifier;
  SIT Sit;
  NUM Atom;
  NUM Value;
  char Type;
} tLITERAL;

typedef struct {
  tLITERAL Consequent;
  struct {
    tLITERAL *Arr;
    NUM NLits;
    QUEUE List;
  } Body;
  unsigned FalseConseq : 1;
} tRULE;

typedef struct {
  NUM NumVal;  
  Atom ***ValAtoms; /* Atom * (ValAtoms[value][situation]) */
  Atom **(PertAtoms[2]); /* 0 = notpertinent, 1 = pertinent */
} tLPATOMS;

typedef struct {
  SIT Sit;
  NUM Val;		/* founded value */
  unsigned isPert:1;	/* pertinence: 1 pertinent - 0 persistent */
  unsigned isKnown:1;    /* known 1 - unknown 0 */
} tVALUE;


typedef struct {
  tVALUE Curr;
  tVALUE *Obs;
  QUEUE Prev;
  tLPATOMS *lpAtoms;  
} tFLUENT;

typedef struct {
  tVALUE Curr;
  QUEUE Prev;
  tLPATOMS *lpAtoms;
} tACTION;


static QUEUE LstRul=qEMPTY;
static NUM NFluents;
static NUM NActions;
static tRULE **Rules=NULL;
static tFLUENT *Fluents=NULL;
static tACTION *Actions=NULL;
Atom *no;

struct tGenericFunctions SMGF = 
  {SMInit,
   SMPrepare,
   SMInitAtom,
   SMTreatRule,
   SMPushSit,
   SMPopSit,
   SMTreatInit,
   SMEndInits,
   SMTreatObs,
   SMTreatAssign,
   SMRestart,
   SMNewTrans,
   SMShowTrans,
   SMGetCurValue,
   SMGetValue,
   SMShowSolution,
   SMSolveGroundQuery};

#define PERTATOM(t,i,v,s) ((t==ACTION)?Actions[i].lpAtoms->PertAtoms[v][s-1]:Fluents[i].lpAtoms->PertAtoms[v][s-1])
#define VALATOM(t,i,v,s) ((t==ACTION)?Actions[i].lpAtoms->ValAtoms[v][s-1]:Fluents[i].lpAtoms->ValAtoms[v][s])

void SMInit() {
}

void SMInitAtom(struct tAtom *a, char )
{
  tLPATOMS *pA;
  NUM i;

  a->data=malloc(sizeof(tLPATOMS));
  pA=(tLPATOMS *)a->data;
  pA->NumVal=a->pfh->RangeSize;
  /* Get memory for all values */
  pA->ValAtoms=(Atom ***) malloc(sizeof(Atom **)*pA->NumVal);
}

static void TreatDep(struct tFunct *pf,struct Elem *e,tRULE *r)
{
  tLITERAL l;

  l.Modifier=pf->Modifier;
  l.Sit=pf->Sit;
  if (l.Modifier==modPERT)
    l.Value=e->N.Sym;
  else
    GetValueNum(pf,e,&l.Value);
  l.Atom=GetAtomNum(pf);
  if (pal_code!=errNONE) return;

  l.Type=pf->Type;

  /* Store literal in the Body list */
  qInsertLast(&(r->Body.List),&l,sizeof(tLITERAL));
}

void SMTreatRule()
{
  unsigned i;
  struct Elem e;
  struct tExpNode Node;
  NUM n;
  tRULE *r;

  /* Create new Rule pattern */
  r=(tRULE *) malloc(sizeof(tRULE));
  r->Body.List=qEMPTY;
  
  /* Treat head */
  if (HeadFunct==NULL)
    r->FalseConseq=1;
  else {
    if (!(HeadFunct->Type==FLUENT 
          && HeadFunct->Modifier==modNONE
	  && HeadFunct->Sit==0)) {
      printf("%d: rule head must be a fluent.\n",yyline); 
      exit(1);
    }  
    Eval(HeadExpr,&Node);
    r->FalseConseq=0;
    e.Type=Node.Type;
    switch (e.Type) {
      case SVALUE : e.N.Sym=Node.d.Sym; break;
      case NVALUE : e.N.Val=Node.d.Val; break;
    }
    r->Consequent.Atom=GetAtomNum(HeadFunct);
    if (pal_code==errBAD_ATOM_ARG) {
	if (OptVal[optWARNINGS]) printf("%d: Warning. %s\n",yyline,pal_errmsg);
	free(r);
	return;
    }

    
    if (GetValueNum(HeadFunct,&e,&n)) {
      r->Consequent.Value=n;
    }
    else {
      if (OptVal[optWARNINGS])
      {
	printf("%d: Warning: head `",yyline);
	showAtom(GetAtomNum(HeadFunct),HeadFunct->Type);
	printf("=");
	showElem(&e);
	printf("' out of range in rule %d. Converted into `false'.\n",
  	       RuleCount);       
      }
      r->FalseConseq=1;
    }
  }

  /* Treat body */
  for (i=0;i<NVars;i++)
    if (VarVec[i].Type!=VAR) {
      TreatDep(VarVec[i].d.Funct,&tup[i],r);

      if (pal_code==errBAD_ATOM_ARG) {
	  unsigned j;      
	  if (OptVal[optWARNINGS]) printf("%d: Warning. %s\n",yyline,pal_errmsg);
	  /* undo dependences */
	  qFree(&(r->Body.List));
	  free(r);
	  pal_code=errBAD_ATOM_ARG;
	  return;
	}      
    }
    
  /* Transform Body List into an array */  
  r->Body.Arr=(tLITERAL *)
    Queue2Array(r->Body.List,sizeof(tLITERAL),&r->Body.NLits);
  qFree(&r->Body.List);

  /* Store rule in List of Rules */
  qInsertLast(&LstRul,&r,sizeof(tRULE *));  

  #ifdef DEBUG
  printf("R%d ",NRules);
  showRule(NVars);
  #endif DEBUG
}


void SMPrepare()
/* Malloc arrays */
{
  static char firsttime=1;  
  NUM i;
  
  
  if (!firsttime) return; else firsttime=0;

  Rules=(tRULE **)Queue2Array(LstRul,sizeof(tRULE *),&i);
  qFree(&LstRul);

  /* malloc action table (calloc sets zeroes) */
  NActions=TabAct.NSym;
  Actions=(tACTION *)calloc(NActions,sizeof(tACTION));
  for (i=0;i<NActions;i++)
    Actions[i].lpAtoms=(tLPATOMS *)((struct tAtom *)stEntry(TabAct,i))->data;
  
  
  /* malloc fluent table */
  NFluents=TabFlu.NSym;
  Fluents=(tFLUENT *)calloc(NFluents,sizeof(tFLUENT));
  for (i=0;i<NFluents;i++)
    Fluents[i].lpAtoms=(tLPATOMS *)((struct tAtom *)stEntry(TabFlu,i))->data;
      
}


void SMRestart()
{
  NUM i;

  Now=0;

  /* init fluent values */  
  for (i=0;i<NFluents;i++) {
    qFree(&Fluents[i].Prev);
    Fluents[i].Curr.Sit=0; /* Now is 0 */
    Fluents[i].Curr.isPert=0;
    Fluents[i].Curr.isKnown=0;
  }
  
  /* init action values */  
  for (i=0;i<NActions;i++) {
    qFree(&Actions[i].Prev);
    Actions[i].Curr.Sit=0; /* Now is 0 */
    Actions[i].Curr.isPert=0;
  }
    
  printf("\nRestart\n");
}

void SMPushSit()
{
  NUM i;

  #ifdef DEBUG
  printf("Push situation %ld\n",Now);
  #endif
  
  Now++;
  
  /* Shift situation and reset marks */

  for(i=0;i<NActions;i++)
  {
    /* if action has been performed, push it */
    if (Actions[i].Curr.isPert)
      qInsertFirst(&Actions[i].Prev,&Actions[i].Curr,sizeof(tVALUE));
    Actions[i].Curr.isPert=0;
    Actions[i].Curr.Sit=Now;
  }
  
  for(i=0;i<NFluents;i++)
  {  
    /* if fluent does not persist or had not prev data, push it */
    if (Fluents[i].Prev==NULL || Fluents[i].Curr.isPert)
      /* push fluent value */
      qInsertFirst(&Fluents[i].Prev,&Fluents[i].Curr,sizeof(tVALUE));
    /* else: it persisted, then data remains unchanged */
        
    Fluents[i].Curr.isKnown=0;
    Fluents[i].Curr.isPert=0;
    Fluents[i].Curr.Sit=Now;
    if (Fluents[i].Obs!=NULL) {free(Fluents[i].Obs); Fluents[i].Obs=NULL;}
  }
}

void SMPopSit()
{
  NUM i;

  Now--;

  #ifdef DEBUG
  printf("Pop situation %ld\n",Now);
  #endif
  
  for(i=0;i<NFluents;i++) {
    if (Fluents[i].Prev==NULL) 
      {printf("PopSit on empty queue fluent ");
       showAtom(i,FLUENT);
       printf("\n");
       exit(0);}

    /* Retrieve previous value */
    memcpy(&Fluents[i].Curr,qFirst(Fluents[i].Prev,tVALUE),
           sizeof(tVALUE));

    if (Fluents[i].Curr.Sit==Now) /* it had not persisted */
      qDeleteFirst(&Fluents[i].Prev);
    else /* it had persisted */
    {
      Fluents[i].Curr.isPert=0;
      Fluents[i].Curr.Sit=Now;
    }    

  }

  for(i=0;i<NActions;i++)
  {
    if (Actions[i].Prev!=NULL && qFirst(Actions[i].Prev,tVALUE)->Sit==Now)
    {
      memcpy(&Actions[i].Curr,qFirst(Actions[i].Prev,tVALUE),
             sizeof(tVALUE));
      qDeleteFirst(&Actions[i].Prev);
    }
    else {
      Actions[i].Curr.isPert=0;
      Actions[i].Curr.Sit=Now;
    }
  }
}


int SMTreatInit(NUM atom,NUM value)
{
  tFLUENT *pFlu;
  
  pFlu=&Fluents[atom];
  if (pFlu->Curr.isKnown && pFlu->Curr.Val!=value) {
    if (verbose) {
      printf("Inconsistence (initial values for fluent `");
      showAtom(atom,FLUENT);
      printf("')\n"); 
    }
    return 0;
  }
  pFlu->Curr.Val=value;
  pFlu->Curr.isKnown=1;

  #ifdef DEBUG
  printf("Init ");
  showAtom(atom,FLUENT);
  printf(":=");
  showAtomVal(atom,value,FLUENT);
  printf("\n");
  #endif  
  
  return 1;
}

void SMEndInits() {
}

int SMTreatObs(NUM ,NUM ,char )
{
  /* Do nothing by now. Only for planning */

  return 1;  
}


int SMTreatAssign(NUM atom,NUM value)
{
  tACTION *pAct;
  
  pAct=&Actions[atom];
  #ifdef DEBUG
  printf("Assign ");
  showAtom(atom,ACTION);
  printf(":=");
  showAtomVal(atom,value,ACTION);
  printf("\n");
  #endif 
  if (pAct->Curr.isPert && pAct->Curr.Val!=value) {
    if (verbose) {
      printf("Inconsistence (values for action `");
      showAtom(atom,ACTION);
      printf("')\n"); 
    }
    return 0;
  }
  pAct->Curr.Val=value;
  pAct->Curr.isPert=1;
  return 1;
}

void ShowValue(tVALUE *pv,NUM i,char Type)
{
  if (pv->isPert) {
    showAtom(i,Type);
    printf(":=");
    showAtomVal(i,pv->Val,Type);
    printf("\n");
  }
}

void SMShowTrans() {
  NUM i;
  printf("%ld)\n",Now);
  for (i=0;i<NActions;i++)
    ShowValue(&Actions[i].Curr,i,ACTION);
  for (i=0;i<NFluents;i++)
    ShowValue(&Fluents[i].Curr,i,FLUENT);
}

/*---------- Routines for smodels ----------*/

void GenLPAtoms(Api *api,NUM n, char Type, SIT length)
{
  tLPATOMS *pA;
  NUM i,v;
  SIT l;
  
  char *AtomName=AtomNum2str(n,Type);

  l=length;      
  if (Type==ACTION) pA=Actions[n].lpAtoms;
  else             {pA=Fluents[n].lpAtoms; l++;}

  /* There are not actions nor pertinence in situation 0 */
  /* 0 means situation 1 for actions and pertinence */

  for (v=0;v<pA->NumVal;v++) {
  
    /* Get memory for all the situations */  
    pA->ValAtoms[v]=(Atom **)malloc(sizeof(Atom *)*l);
    
    for (i=0;i<l;i++) {
      pA->ValAtoms[v][i]=(*api).new_atom();
      if (fTxtProgram)
      {
	char *s=NULL,*t=NULL,r[20];
	
	s=strCopy("h(");
        s=strAppend(s,AtomName);
	s=strAppend(s,",");
	t=AtomVal2str(n,v,Type);
	s=strAppend(s,t);
	s=strAppend(s,",");
	sprintf(r,"%d",(Type==ACTION)?i+1:i);
	s=strAppend(s,r);
	s=strAppend(s,")");
	(*api).set_name(pA->ValAtoms[v][i],s);
	free(s);
	free(t);
      }
    }
  }
  
  for (v=0;v<2;v++) {
    /* Get memory for all the situations */  
    pA->PertAtoms[v]=(Atom **)malloc(sizeof(Atom *)*length);
    
    for (i=0;i<length;i++) {
      pA->PertAtoms[v][i]=(*api).new_atom();
      if (fTxtProgram)
      {
	char *s=NULL,r[20];
	
	sprintf(r,"%d",i+1);
        s=strCopy("p(");
	s=strAppend(s,AtomName);
	s=(v==0)?strAppend(s,",n,"):strAppend(s,",p,");
	s=strAppend(s,r);
	s=strAppend(s,")");
	(*api).set_name(pA->PertAtoms[v][i],s);
	free(s);
      }
    }
  }

  #ifdef DEBUG
  free(AtomName);
  #endif          
}

void FreeLPAtoms(NUM n, char Type) {
  tLPATOMS *pA;
  NUM i,v;
  
  pA=(Type==ACTION)?Actions[n].lpAtoms:Fluents[n].lpAtoms;

  for (v=0;v<pA->NumVal;v++)
    free(pA->ValAtoms[v]);
  
  for (v=0;v<2;v++)
    free(pA->PertAtoms[v]);
}

void GenLPBody(Api *api,tRULE *r, SIT sit) 
{
  NUM j;
  tLITERAL *l;
  SIT s;
  tLPATOMS *pA=NULL;
  void *table;

  /* body literals */
  for (j=0;j<r->Body.NLits;j++) {
    l=&(r->Body.Arr[j]);
    s = sit + l->Sit;
    
    /* pertinence and actions in the initial situation...*/
    if ((l->Type==ACTION||l->Modifier==modPERT)&&s==0)  
      (*api).add_body(no,true); /*...  are always false */
    else
      if (l->Modifier==modNONE)
	(*api).add_body(VALATOM(l->Type,l->Atom,l->Value,s),true);
      else 
	(*api).add_body(PERTATOM(l->Type,l->Atom,l->Value,s),true);
  }
}

void GenRulePair(Api *api,tRULE *r,char WithPert, NUM PertLit, SIT sit) 
{
  Atom *p;
  
  (*api).begin_rule (BASICRULE);
  if (r->FalseConseq)
    (*api).add_head(no);
  else
    (*api).add_head(
        VALATOM(FLUENT,r->Consequent.Atom,r->Consequent.Value,sit));
	
  GenLPBody(api,r,sit);
  if (WithPert) {
    /* pertinence literal */
    p=PERTATOM(FLUENT,r->Body.Arr[PertLit].Atom,TRUE_NSym,sit);
    (*api).add_body(p,true);      
  }
  (*api).end_rule ();
  if (!r->FalseConseq) /* add rule for pertinence for the head */
  {
    (*api).begin_rule (BASICRULE);
    (*api).add_head(
        PERTATOM(FLUENT,r->Consequent.Atom,TRUE_NSym,sit));
    GenLPBody(api,r,sit);
    if (WithPert)
      (*api).add_body(p,true);      
    (*api).end_rule ();
  }
}

void GenLPRules(Api *api,NUM nr,SIT sit)
{
  NUM i;
  tRULE *r=Rules[nr];    
  
  /* look whether there is some current-action-literal */
  for (i=0; i<r->Body.NLits &&
            !(r->Body.Arr[i].Type==ACTION && r->Body.Arr[i].Sit==0 &&
	     !(r->Body.Arr[i].Modifier==modPERT && 
	       r->Body.Arr[i].Value==FALSE_NSym)) ; i++);

  /* if so, generate just one rule pair */
  if (i<r->Body.NLits)
    GenRulePair(api,r,0,0,sit);
  else
    /* no actions: generate two rules for each pertinent-literal */

    for (i=0;i<r->Body.NLits;i++)
      if (r->Body.Arr[i].Sit==0)       /* the literal yields pertinence*/
        GenRulePair(api,r,1,i,sit);
}


void MinimizePert(Api *api,char Type,NUM i,SIT sit) {
  /* Pertinence minimization */    
  /* pert(p) :- not negpert(p) */
  (*api).begin_rule (BASICRULE);
  (*api).add_head(PERTATOM(Type,i,FALSE_NSym,sit));
  (*api).add_body(PERTATOM(Type,i,TRUE_NSym, sit),false);
  (*api).end_rule ();
}

void GenLPAxioms(Api *api,SIT length) {
  NUM i,v,w,numval;
  SIT s;  
    
  for (s=1;s<=length;s++) {
    for (i=0;i<NFluents;i++) {
      numval=Fluents[i].lpAtoms->NumVal;

      /* Uniqueness of value */
      /* *** investigate if can be done with weight rule */
      for (v=0;v<numval;v++) 
	for (w=v+1;w<numval;w++)
	{
	  /*
	     false :- holds(p,v),holds(p,w)
	     note:the consistence in initial state 
	     is externally guaranteed
	  */
	  (*api).begin_rule (BASICRULE); 
	  (*api).add_head(no);
	  (*api).add_body(VALATOM(FLUENT,i,v,s),true);
	  (*api).add_body(VALATOM(FLUENT,i,w,s),true);
	  (*api).end_rule ();
	}
      MinimizePert(api,FLUENT,i,s);

      /* For each value */
      for (v=0;v<numval;v++) {
	/* Inertia: */
	/* holds(p,v) :- negpert(p), holds(prev(p),v) */
	(*api).begin_rule (BASICRULE);
	(*api).add_head(VALATOM(FLUENT,i,v,s));
	(*api).add_body(PERTATOM(FLUENT,i,FALSE_NSym,s),true);
	(*api).add_body(VALATOM(FLUENT,i,v,s-1),true);
	(*api).end_rule ();
      }      
    }

    for (i=0;i<NActions;i++)  {
      numval=Actions[i].lpAtoms->NumVal;
      /* Uniqueness of value */
      /* controlled externally (for temporal projection) */
      /* guaranteed by choice rules (for planning) */
      
      MinimizePert(api,ACTION,i,s);

      /* For each value */
      for (v=0;v<numval;v++) {
	/* Action occurrence implies pertinence of action */
	/* pert(a) :- holds(a,v) */
	(*api).begin_rule (BASICRULE);
	(*api).add_head(PERTATOM(ACTION,i,TRUE_NSym,s));
	(*api).add_body(VALATOM(ACTION,i,v,s),true);
	(*api).end_rule ();
      }
    }
  }
}


void GenChoiceRules(Api *api,SIT length) {
  /* *** Try to use a weight rule instead */
  NUM i,j,v,w;
  SIT s;
  
  for (s=1;s<=length;s++)
    for (i=0;i<NActions;i++)
      for (v=0;v<Actions[i].lpAtoms->NumVal;v++)
      {
	(*api).begin_rule (BASICRULE);
	(*api).add_head(VALATOM(ACTION,i,v,s));
	for (w=0;w<Actions[i].lpAtoms->NumVal;w++)
          if (v!=w)
	    (*api).add_body(VALATOM(ACTION,i,w,s),false);
	(*api).add_body(PERTATOM(ACTION,i,FALSE_NSym,s),false);
	(*api).end_rule ();
      }
  if (!OptVal[optCONCURRENT])
    for (s=1;s<=length;s++) {
      for (i=0;i<NActions;i++)
	for (v=0;v<Actions[i].lpAtoms->NumVal;v++)
          for (j=i+1;j<NActions;j++)
	    for (w=0;w<Actions[j].lpAtoms->NumVal;w++) {
	      (*api).begin_rule (BASICRULE);
	      (*api).add_head(no);
	      (*api).add_body(VALATOM(ACTION,i,v,s),true);
	      (*api).add_body(VALATOM(ACTION,j,w,s),true);
	      (*api).end_rule ();	  
	    }
      (*api).begin_rule (BASICRULE);
      (*api).add_head(no);
      for (i=0;i<NActions;i++)
	(*api).add_body(PERTATOM(ACTION,i,FALSE_NSym,s),true);
      (*api).end_rule ();	  
    }
}

int SMNewTrans()
{
  NUM i,v,numval;
  char ok=1;
  tVALUE *pv;
  Smodels smodels;
  Api api (&smodels.program);
  
  #ifdef DEBUG
  printf("--- NEW TRANS Now=%ld ---\n",Now);
  #endif


  /* Construct the smodels program */  
  api.remember ();
  no=api.new_atom();
  api.set_name(no,"no");
  api.set_compute(no,false);

  for (i=0;i<NActions;i++) GenLPAtoms(&api,i,ACTION,1);
  for (i=0;i<NFluents;i++) GenLPAtoms(&api,i,FLUENT,1);
  for (i=0;i<NRules;i++) GenLPRules(&api,i,1);
  GenLPAxioms(&api,1);

  /* Add facts for initial state */
  for (i=0;i<NFluents;i++) {
    pv=qFirst(Fluents[i].Prev,tVALUE);
    if (pv->isKnown) {
      api.begin_rule(BASICRULE);
      api.add_head(VALATOM(FLUENT,i,pv->Val,0));
      api.end_rule();
    }
  }
  /* Add facts for action occurrence */
  for (i=0;i<NActions;i++) 
    if (Actions[i].Curr.isPert) {
      api.begin_rule(BASICRULE);
      api.add_head(VALATOM(ACTION,i,Actions[i].Curr.Val,1));
      api.end_rule();
    }
  
  api.done ();
  if (fTxtProgram) {
    FILE *out=stdout;
    stdout=fTxtProgram;
    smodels.program.print ();
    stdout=out;
  }

  smodels.init ();
  if (smodels.model ()) {
    #ifdef DEBUG
    smodels.printAnswer (); 
    #endif
    for (i=0;i<NFluents;i++) {
      if (PERTATOM(FLUENT,i,TRUE_NSym,1)->Bpos) {
        Fluents[i].Curr.isPert=1;
        numval=Fluents[i].lpAtoms->NumVal;
        for(v=0;v<numval && VALATOM(FLUENT,i,v,1)->Bneg;v++);
	if (v>=numval) {
	  printf("Error: no value obtained for a pertinent fluent.\n");
	  exit(1);
	}	
        Fluents[i].Curr.isKnown=1;
        Fluents[i].Curr.Val=v;
      }
    }
        
    if (smodels.model ())
      printf("Several models were found.\n");
  }
  else {
    ok=0; 
    if (verbose)
      printf("No models found.\n"); 
  }
  
  for (i=0;i<NActions;i++) FreeLPAtoms(i,ACTION);
  for (i=0;i<NFluents;i++) FreeLPAtoms(i,FLUENT);
  
  return ok;
}


int SMGetCurValue(struct tVarLstNode *pn,struct Elem *pe)
{
  struct tFunctHeader *pf;
  NUM n;
  tVALUE *Prev;
  tVALUE *PrevA;

  if (Fluents==NULL) {
    printf("%d: ",yyline); 
    printf("Cannot perform `query' before an `initially' clause.\n");
    exit(1);
  }

  n=GetAtomNum(pn->d.Funct);
  if (pal_code!=errNONE) return 0;
  switch (pn->Type) {
    case ACTION:
      switch(pn->d.Funct->Modifier) {
        case modNONE:
	  if (pn->d.Funct->Sit==0) {
	    if (!Actions[n].Curr.isPert) return 0;
	    pf=((struct tAtom *)stEntry(TabAct,n))->pfh;
	    memcpy(pe,&pf->Range[Actions[n].Curr.Val],sizeof(struct Elem));
	  }
	  else {
	    if (Actions[n].Prev==NULL) return 0;
	    PrevA=qFirst(Actions[n].Prev,tVALUE);
	    if (PrevA->Sit!=Now-1) return 0;
	    pf=((struct tAtom *)stEntry(TabAct,n))->pfh;
	    memcpy(pe,&pf->Range[PrevA->Val],sizeof(struct Elem));
	  }
	  break;
        case modPERT:
	  pe->Type=SVALUE;
	  if (pn->d.Funct->Sit==0) {
	    if (Actions[n].Curr.isPert) 
	      pe->N.Sym=TRUE_NSym;
	    else
	      pe->N.Sym=FALSE_NSym;
	  }
	  else {
	    if (Actions[n].Prev==NULL)
	      pe->N.Sym=FALSE_NSym;
	    else {
	      PrevA=qFirst(Actions[n].Prev,tVALUE);
	      if (PrevA->Sit==Now-1) pe->N.Sym=TRUE_NSym;
	      else pe->N.Sym=FALSE_NSym;
	    }
	  }
	  break;
      }
      break;
    case FLUENT:  
      switch(pn->d.Funct->Modifier) {
        case modNONE:
	  if (pn->d.Funct->Sit==0) {
	    if (!Fluents[n].Curr.isKnown) return 0;
	    pf=((struct tAtom *)stEntry(TabFlu,n))->pfh;
	    memcpy(pe,&pf->Range[Fluents[n].Curr.Val],sizeof(struct Elem));
	  }
	  else {
	    if (qIsEmpty(Fluents[n].Prev)) return 0;
            Prev=qFirst(Fluents[n].Prev,tVALUE);
	    if (!Prev->isKnown) return 0;
	    pf=((struct tAtom *)stEntry(TabFlu,n))->pfh;
	    memcpy(pe,&pf->Range[Prev->Val],sizeof(struct Elem));
	  }
	  break;
        case modPERT:
	  if (pn->d.Funct->Sit==0) {
	    pe->Type=SVALUE;
	    if (Fluents[n].Curr.isPert) 
	      pe->N.Sym=TRUE_NSym;
	    else
	      pe->N.Sym=FALSE_NSym;
	  }
	  else {
	    if (qIsEmpty(Fluents[n].Prev)) return 0;
            Prev=qFirst(Fluents[n].Prev,tVALUE);
	    pe->Type=SVALUE;
	    if (Prev->Sit==Now-1 && Prev->isPert) pe->N.Sym=TRUE_NSym;
	    else pe->N.Sym=FALSE_NSym;	    
	  }
	  break;
      }
  }
  return 1;
}

int SMGetValue(NUM n,char Type,NUM *v,char *isPert)
/* returns 0 if fluent unknown or Fluents is NULL
   returns 1 otherwise
*/
{
  if (Fluents==NULL) return 0;

  switch(Type) {
    case ACTION:
      *isPert=Actions[n].Curr.isPert;
      *v=Actions[n].Curr.Val;
      return 1;

    case FLUENT:  
      if (!Fluents[n].Curr.isKnown) return 0;
      *isPert=Fluents[n].Curr.isPert;
      *v=Fluents[n].Curr.Val;
      return 1;
  }
}

void SMShowSolution()
{
  NUM i,v,numval;
  SIT s;
  char first=1;  
  
  NSol++;
  if(ExprSit>1)
    printf("\nSolution %d:\n",NSol);
  for(i=0;i<NVars;i++)
    if (VarVec[i].Type==VAR)
    {
      if (first) first=0; else printf(",");
      printf("%s",stEntry(SymTables[VAR],VarVec[i].d.Sym));
      printf("=");
      showElem(&tup[i]);
    }
  if (!first) printf("\n");

  if (ExprSit<=1) return; /* no temporal query */

  for (s=1;s<ExprSit;s++) {
    printf("%d)\n",s+IniSit);
    for (i=0;i<NActions;i++)
      if (PERTATOM(ACTION,i,TRUE_NSym,s)->Bpos) {
        numval=Actions[i].lpAtoms->NumVal;
        for(v=0;v<numval && VALATOM(ACTION,i,v,s)->Bneg;v++);
	if (v>=numval) {
	  printf("Error: no value obtained for a pertinent action.\n");
	  exit(1);
	}	
        showAtom(i,ACTION);
	printf(":=");
	showAtomVal(i,v,ACTION);
	printf("\n");
      }

    for (i=0;i<NFluents;i++)
      if (PERTATOM(FLUENT,i,TRUE_NSym,s)->Bpos) {
        numval=Fluents[i].lpAtoms->NumVal;
        for(v=0;v<numval && VALATOM(FLUENT,i,v,s)->Bneg;v++);
	if (v>=numval) {
	  printf("Error: no value obtained for a pertinent fluent.\n");
	  exit(1);
	}	
        showAtom(i,FLUENT);
	printf(":=");
	showAtomVal(i,v,FLUENT);
	printf("\n");
      }
  }

}

void SMSolveGroundQuery() {
  Smodels smodels;
  Api api (&smodels.program);
  NUM i,n,v;
  SIT length=ExprSit-1,s;

  #ifdef DEBUG
  printf("Solve query-------\n");
  #endif

  if (ExprSit==1) {
    GF->ShowSolution();
    return;
  }

  /* Construct the smodels program */  
  api.remember ();
  no=api.new_atom();
  api.set_name(no,"no");
  api.set_compute(no,false);

  for (i=0;i<NActions;i++) GenLPAtoms(&api,i,ACTION,length);
  for (i=0;i<NFluents;i++) GenLPAtoms(&api,i,FLUENT,length);
  for (s=1;s<=length;s++)
    for (i=0;i<NRules;i++) 
      GenLPRules(&api,i,s);
  GenLPAxioms(&api,length);
  GenChoiceRules(&api,length);

  /* Add facts for initial state */
  for (i=0;i<NFluents;i++) {
    if (Fluents[i].Curr.isKnown) {
      api.begin_rule(BASICRULE);
      api.add_head(VALATOM(FLUENT,i,Fluents[i].Curr.Val,0));
      api.end_rule();
    }
  }
  
  /* Add observations from the query case */
  for (i=0;i<QueryVec[ExprSit-1].NVars;i++)
    if (VarVec[i].Type!=VAR) {
      n=GetAtomNum(VarVec[i].d.Funct);
      switch(VarVec[i].d.Funct->Modifier) {
        case modPERT:
	  api.set_compute(
	      PERTATOM(VarVec[i].Type,n,tup[i].N.Sym,VarVec[i].d.Funct->Sit),
	      true);
	  break;
	  
	case modNONE:
	  GetValueNum(VarVec[i].d.Funct,&tup[i],&v);
	  api.set_compute(
	      VALATOM(VarVec[i].Type,n,v,VarVec[i].d.Funct->Sit),
	      true);
      }
    }

  api.done ();
  #ifdef DEBUG
  smodels.program.print ();
  #endif

  smodels.init ();
  while (smodels.model ()) {
    #ifdef DEBUG
    smodels.program.print ();
    #endif
    GF->ShowSolution();    
    if(OptVal[optSOLUTIONS] && NSol>=OptVal[optSOLUTIONS]) 
      break;
  
  }

  for (i=0;i<NActions;i++) FreeLPAtoms(i,ACTION);
  for (i=0;i<NFluents;i++) FreeLPAtoms(i,FLUENT);

}
