#include "all.h"

/*#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "queue.h"
#include "btree.h"
#include "binvec.h"
#include "icode.h"
#include "wfs.h"
#include "sstack.h"
*/

typedef struct {
  char Modifier;
  NUM Atom;
  NUM Value;
} tLITERAL;

typedef struct {
  NUM Rule;
  NUM Value;
  char Modifier;
  SIT Sit;
  char Done;  /* dependence done, at the loop detection */
} tFWDDEP;

typedef struct {
  QUEUE deps;
  NUM NRules;
} tATOMDATA;

typedef struct {
  tLITERAL Consequent;
  NUM NLits;    /* Number of literals */  
  struct { 
    NUM NRemLits; 
    char isPert;
  } N; /* Remaining literals (nonunfounded) */
  struct { 
    NUM NRemLits; 
    char isPert;
  } F; /* Remaining literals (founded) */
  NUM NPerts;   /* Number of literals yielding pertinence */
  NUM NRemPerts;
  unsigned FalseConseq: 1;
  unsigned Unfounded: 1;
} tRULE;


typedef struct {
  NUM Sit;
  struct {
    BINVEC Val;
    unsigned isPert:1;    
  } N;
  struct {
    NUM Val;    
    unsigned isPert:1;    
  } F;
  unsigned isKnown:1;    
} tFLUENTVAL;

typedef struct {
  NUM Sit;
  NUM Val;
  unsigned isPert:1;
} tACTIONVAL;

typedef struct {
  NUM NumVal;  
  tFLUENTVAL Curr;
  tFLUENTVAL *Obs;
  QUEUE Prev;
  unsigned NRules;	/* num. of rules whose effect is this fluent */
  unsigned NRemRules;	/* num. of remaining rules */
} tFLUENT;

typedef struct {
  tACTIONVAL Curr;
  QUEUE Prev;
} tACTION;

#define APPLICABLE(r) (!Rules[r]->F.NRemLits && Rules[r]->F.isPert)
#define N_APPLICABLE(r) (!Rules[r]->N.NRemLits && Rules[r]->N.isPert)
#define DEPQUEUE(TAB,N) (((tATOMDATA *)((struct tAtom *)stEntry((TAB),(N)))->data)->deps)
#define NRULES(TAB,N) (((tATOMDATA *)((struct tAtom *)stEntry((TAB),(N)))->data)->NRules)
#define DECREMENT(X) {if (X>0) (X)--;}

QUEUE LstRul=qEMPTY;
struct tGenericFunctions WFGF = 
  {WFInit,
   WFPrepare,
   WFInitAtom,
   WFTreatRule,
   WFPushSit,
   WFPopSit,
   WFTreatInit,
   WFEndInits,
   WFTreatObs,
   WFTreatAssign,
   WFRestart,
   WFNewTrans,
   WFShowTrans,
   WFGetCurValue,
   WFGetValue,
   WFShowSolution,
   SolveGroundQuery};

/* Local variables */
static NUM NFluents;
static NUM NActions;
static tRULE **Rules=NULL;
static QUEUE *ActionDeps=NULL;
static QUEUE *FluentDeps=NULL;
static tFLUENT *Fluents=NULL;
static tACTION *Actions=NULL;

void UpdateCurRules(NUM r) ;
int SolvePrevDep(QUEUE q,void *data);
void FireDep(tFWDDEP *dep,char isPert);
int SolveActDep(QUEUE q,void *data);
int FireRule();
int Persistence();
int SolveFluDep(QUEUE q,void *data);

int ResetDone(QUEUE q,void *d);
void NFireDep(tFWDDEP *dep,char isPert);
int NSolveFluDep(QUEUE q,void *data);
void NFireRule();
void LoopDetection();

static SSTACK CurRules,CurFluents;

#ifdef DEBUG
void PrintSS(char *name,SSTACK s)
{ NUM i;

  printf("%s ",name);
  if (ssISEMPTY(s)) printf("EMPTY");
  else
    for (i=ssTOP(s)+1;i!=0;i=s[i].next)
      printf("%d ",i-1);
  printf("\n");
}
#endif

void WFInit() {
}

void WFInitAtom(struct tAtom *a, char Type)
{
  a->data=(void *)malloc(sizeof(tATOMDATA));
  ((tATOMDATA *)(a->data))->deps=NULL;
  ((tATOMDATA *)(a->data))->NRules=0;
}

static void TreatDep(struct tFunct *pf,struct Elem *e,tRULE *r)
{
  tFWDDEP dep;
  NUM n;

  if (pf->Type==FLUENT) r->NLits++;
  if (pf->Sit==0)       r->NPerts++;
  /* Create forward dependence */
  dep.Rule=NRules;
  dep.Modifier=pf->Modifier;
  dep.Sit=pf->Sit;
  if (dep.Modifier==modPERT)
    dep.Value=e->N.Sym;
  else
    GetValueNum(pf,e,&dep.Value);

  /* Store dependence in the corresponding atom */
  n=GetAtomNum(pf);
  if (pal_code!=errNONE) return;
  switch (pf->Type) {
    case ACTION: qInsertFirst(&DEPQUEUE(TabAct,n),&dep,sizeof(tFWDDEP)); break;
    case FLUENT: qInsertFirst(&DEPQUEUE(TabFlu,n),&dep,sizeof(tFWDDEP)); break;
  }
}

static void UndoDep(struct tFunct *pf)
{
  NUM n;

  /* Delete dependence from the corresponding atom */
  n=GetAtomNum(pf);
  switch (pf->Type) {
    case ACTION: qDeleteFirst(&DEPQUEUE(TabAct,n)); break;
    case FLUENT: qDeleteFirst(&DEPQUEUE(TabFlu,n)); break;
  }
}

void WFTreatRule()
{
  unsigned i;
  struct Elem e;
  struct tExpNode Node;
  NUM n;
  tRULE *r;

  /* Create new Rule */
  r=(tRULE *) malloc(sizeof(tRULE));
  r->Unfounded=0;
  r->F.isPert=0;
  r->NLits=0;
  r->NPerts=0;
  
  /* Treat head */
  if (HeadFunct==NULL)
    r->FalseConseq=1;
  else {
    if (!(HeadFunct->Type==FLUENT 
          && HeadFunct->Modifier==modNONE
	  && HeadFunct->Sit==0)) {
      printf("%d: rule head must be a fluent.\n",yyline); 
      exit(1);
    }  
    Eval(HeadExpr,&Node);
    r->FalseConseq=0;
    e.Type=Node.Type;
    switch (e.Type) {
      case SVALUE : e.N.Sym=Node.d.Sym; break;
      case NVALUE : e.N.Val=Node.d.Val; break;
    }
    r->Consequent.Atom=GetAtomNum(HeadFunct);
    if (pal_code==errBAD_ATOM_ARG) {
      if (OptVal[optWARNINGS]) printf("%d: Warning. %s\n",yyline,pal_errmsg);
      free(r);
      return;
    }
    
    if (GetValueNum(HeadFunct,&e,&n)) {
      r->Consequent.Value=n;
    }
    else {
      if (OptVal[optWARNINGS])
      {
	printf("Warning: head `");
	showAtom(r->Consequent.Atom,HeadFunct->Type);
	printf(":=");
	showElem(&e);
	printf("' out of range in rule %d. Converted into `false'.\n",
  	       RuleCount,yyline);       
      }
      r->FalseConseq=1;
    }
  }


  if (!r->FalseConseq)
    NRULES(TabFlu,r->Consequent.Atom)++;
    
  /* Treat body: store dependences in Deps queue */
  for (i=0;i<NVars;i++)
    if (VarVec[i].Type!=VAR) {
      TreatDep(VarVec[i].d.Funct,&tup[i],r);

      if (pal_code==errBAD_ATOM_ARG) {
	unsigned j;      
	if (OptVal[optWARNINGS]) printf("%d: Warning. %s\n",yyline,pal_errmsg);
	/* undo dependences */
	for (j=0;j<i;j++)
	  if (VarVec[j].Type!=VAR) UndoDep(VarVec[j].d.Funct);	
	free(r);
	pal_code=errBAD_ATOM_ARG;
	return;
      }

    }
    
  r->F.NRemLits=r->NLits;
  r->NRemPerts=r->NPerts;

  /* Store rule in List of Rules */
  qInsertLast(&LstRul,&r,sizeof(tRULE *)); 
  
  #ifdef DEBUG
  printf("R%d ",NRules);
  showRule(NVars);
  #endif DEBUG
  return; 

}

void WFPrepare()
/* Malloc arrays */
{
  static char firsttime=1;  
  NUM i;
  
  if (!firsttime) return;
  Rules=(tRULE **)Queue2Array(LstRul,sizeof(tRULE *),&i);
  qFree(&LstRul);

  NFluents=TabFlu.NSym;
  NActions=TabAct.NSym;

  /* malloc action table (calloc sets zeroes) */
  Actions=(tACTION *)calloc(NActions,sizeof(tACTION));
  
  /* malloc fluent table */
  Fluents=(tFLUENT *)calloc(NFluents,sizeof(tFLUENT));
  for (i=0;i<NFluents;i++) {
    Fluents[i].NumVal=((struct tAtom *)stEntry(TabFlu,i))->pfh->RangeSize;
    Fluents[i].Curr.N.Val=bvNew(Fluents[i].NumVal);
    Fluents[i].Curr.N.isPert=1;
    Fluents[i].Curr.F.isPert=0;
    Fluents[i].Prev=NULL;
    Fluents[i].NRules= Fluents[i].NRemRules=NRULES(TabFlu,i);
  }
  firsttime=0;

  /* init fluent dependences */
  FluentDeps=(QUEUE *)malloc(sizeof(QUEUE)*NFluents);
  for (i=0;i<NFluents;i++) {
    FluentDeps[i]=DEPQUEUE(TabFlu,i);
    free(((struct tAtom *)stEntry(TabFlu,i))->data);
    ((struct tAtom *)stEntry(TabFlu,i))->data=NULL;
  }

  /* init action dependences */
  ActionDeps=(QUEUE *)malloc(sizeof(QUEUE)*NActions);
  for (i=0;i<NActions;i++) {
    ActionDeps[i]=DEPQUEUE(TabAct,i);
    free(((struct tAtom *)stEntry(TabAct,i))->data);
    ((struct tAtom *)stEntry(TabAct,i))->data=NULL;
  }

}

void WFRestart()
{
  NUM i,v;
  tFLUENTVAL *Prev;

  Now=0;

  /* init fluent values */  
  for (i=0;i<NFluents;i++) {
    for (;qIsNotEmpty(Fluents[i].Prev);qDeleteFirst(&Fluents[i].Prev))
      free(qFirst(Fluents[i].Prev,tFLUENTVAL)->N.Val);
    Fluents[i].Curr.Sit=0; /* Now is 0 */
    Fluents[i].Curr.N.isPert=1;
    Fluents[i].Curr.F.isPert=0;
    Fluents[i].Curr.isKnown=0;
    for (v=0;v<Fluents[i].NumVal;v++)
      bvSet(Fluents[i].Curr.N.Val,v,0);
    Fluents[i].NRemRules=Fluents[i].NRules;
  }
  
  /* init action values */  
  for (i=0;i<NActions;i++) {
    qFree(&Actions[i].Prev);
    Actions[i].Curr.Sit=0; /* Now is 0 */
    Actions[i].Curr.isPert=0;
  }
    
  printf("\nRestart\n");
}

void showbinvec(BINVEC bv,unsigned size)
{
  unsigned i;
  
  printf(" ");
  for(i=0;i<size;i++)
    printf("%d",(bvVal(bv,i)!=0));
  printf("\n");
}

void WFPushSit()
{
  NUM i;

  #ifdef DEBUG
  printf("Push situation %d\n",Now);
  #endif
  
  Now++;
  
  /* Shift situation and reset marks */

  for(i=0;i<NActions;i++)
  {
    /* if action has been performed, push it */
    if (Actions[i].Curr.isPert)
      qInsertFirst(&Actions[i].Prev,&Actions[i].Curr,sizeof(tACTIONVAL));
    Actions[i].Curr.isPert=0;
    Actions[i].Curr.Sit=Now;
  }
  
  for(i=0;i<NFluents;i++)
  {  
    /* if fluent does not persist or had not prev data, push it */
    if (Fluents[i].Prev==NULL || Fluents[i].Curr.N.isPert) {
      /* push fluent value */
      qInsertFirst(&Fluents[i].Prev,&Fluents[i].Curr,sizeof(tFLUENTVAL));
      /* make a copy of NVal marks so they can be modified */
      qFirst(Fluents[i].Prev,tFLUENTVAL)->N.Val=
             bvCopy(Fluents[i].Curr.N.Val,Fluents[i].NumVal);
    }
    /* else: it persisted, then data remains unchanged */
        
    Fluents[i].Curr.N.isPert=1;
    Fluents[i].Curr.F.isPert=0;
    Fluents[i].Curr.isKnown=0;
    Fluents[i].Curr.Sit=Now;
    if (Fluents[i].Obs!=NULL) {free(Fluents[i].Obs); Fluents[i].Obs=NULL;}
    /* Reset counters */
    Fluents[i].NRemRules=Fluents[i].NRules;
  }
  
  /* Reset rule marks */
  for (i=0;i<NRules;i++) {
    Rules[i]->F.NRemLits=Rules[i]->NLits;
    Rules[i]->NRemPerts=Rules[i]->NPerts;
    Rules[i]->F.isPert=0;
    Rules[i]->Unfounded=0;
  }    
}


void WFPopSit()
{
  NUM i,v;
  BINVEC auxNVal;

  Now--;

  #ifdef DEBUG
  printf("Pop situation %d\n",Now);
  #endif
  
  for(i=0;i<NFluents;i++) {
    if (Fluents[i].Prev==NULL) 
      {printf("PopSit on empty queue fluent ");
       showAtom(i,FLUENT);
       printf("\n");
       exit(0);}

    free(Fluents[i].Curr.N.Val);

    /* Retrieve previous value (this overrides NVal pointer) */
    memcpy(&Fluents[i].Curr,qFirst(Fluents[i].Prev,tFLUENTVAL),
           sizeof(tFLUENTVAL));

    if (Fluents[i].Curr.Sit==Now) /* it had not persisted */
      qDeleteFirst(&Fluents[i].Prev);
    else /* it had persisted */
    {
      Fluents[i].Curr.N.isPert=0;
      Fluents[i].Curr.F.isPert=0;
      Fluents[i].Curr.Sit=Now;
      /* Make a copy of N.Val so the one in prev is a different one */
      Fluents[i].Curr.N.Val=bvCopy(Fluents[i].Curr.N.Val,Fluents[i].NumVal);
    }    

  }

  for(i=0;i<NActions;i++)
  {
    if (Actions[i].Prev!=NULL && qFirst(Actions[i].Prev,tACTIONVAL)->Sit==Now)
    {
      memcpy(&Actions[i].Curr,qFirst(Actions[i].Prev,tACTIONVAL),
             sizeof(tACTIONVAL));
      qDeleteFirst(&Actions[i].Prev);
    }
    else {
      Actions[i].Curr.isPert=0;
      Actions[i].Curr.Sit=Now;
    }
  }
}

int WFTreatInit(NUM atom,NUM value)
{
  tFLUENT *pFlu;
  
  pFlu=&Fluents[atom];
  if (pFlu->Curr.isKnown && pFlu->Curr.F.Val!=value) {
    if (verbose) {
      printf("Inconsistence (initial values for fluent `");
      showAtom(atom,FLUENT);
      printf("')\n"); 
    }
    return 0;
  }
  bvSet(pFlu->Curr.N.Val,value,1);
  pFlu->Curr.N.isPert=0;
  pFlu->Curr.F.Val=value;
  pFlu->Curr.F.isPert=0;
  pFlu->Curr.isKnown=1;
  
  #ifdef DEBUG
  printf("Init ");
  showAtom(atom,FLUENT);
  printf(":=");
  showAtomVal(atom,value,FLUENT);
  printf("\n");
  #endif  
  
  return 1;
}

void WFEndInits() {
}

int WFTreatObs(NUM atom,NUM value,char modifier)
{
  tFLUENT *pFlu;
  
  #ifdef DEBUG
  printf("TreatObs\n");
  #endif  
  pFlu=&Fluents[atom];
  if (pFlu->Obs==NULL) {
    pFlu->Obs=(tFLUENTVAL *)calloc(1,sizeof(tFLUENTVAL));
    pFlu->Obs->N.isPert=1;
    /* pFlu->Obs->F.isPert=0; */
  }
  switch (modifier) {
    case modNONE:
      if (pFlu->Obs->isKnown && pFlu->Obs->F.Val!=value) return 0;
      pFlu->Obs->isKnown=1;
      pFlu->Obs->F.Val=value;
      break;
    case modPERT:
      if (value==0 && pFlu->Obs->F.isPert || value==1 && !pFlu->Obs->N.isPert) return 0;
      pFlu->Obs->F.isPert=pFlu->Obs->N.isPert=value;
      break;
  }
  return 1;  
}


int WFTreatAssign(NUM atom,NUM value)
{
  tACTION *pAct;
  
  pAct=&Actions[atom];
  #ifdef DEBUG
  printf("Assign ");
  showAtom(atom,ACTION);
  printf(":=");
  showAtomVal(atom,value,ACTION);
  printf("\n");
  #endif 
  if (pAct->Curr.isPert && pAct->Curr.Val!=value) {
    if (verbose) {
      printf("Inconsistence (values for action `");
      showAtom(atom,ACTION);
      printf("')\n"); 
    }
    return 0;
  }
  pAct->Curr.Val=value;
  pAct->Curr.isPert=1;
  return 1;
}

void ShowFluentVal(tFLUENTVAL *pfv,NUM i)
{
  NUM v;
  if (!pfv->F.isPert && pfv->N.isPert) {      
    showAtom(i,FLUENT);
    printf(":=?   Unfounded values={ ");
    for (v=0;v<Fluents[i].NumVal;v++)
      if (!bvVal(pfv->N.Val,v)) 
	{showAtomVal(i,v,FLUENT); printf(" ");}
    printf("}\n");      
  }
  else
    if (pfv->F.isPert) {
      showAtom(i,FLUENT);
      printf(":=");
      showAtomVal(i,pfv->F.Val,FLUENT);
      printf("\n");
    }
    #ifdef DEBUG    
    else {
      printf("[");
      showAtom(i,FLUENT);
      printf(":=");
      if (pfv->isKnown)
        showAtomVal(i,pfv->F.Val,FLUENT);
      else
        printf("?");
      printf("]\n");      
    }
    #endif
}

void ShowActionVal(tACTIONVAL *pav,NUM i)
{
  if (pav->isPert) {
    showAtom(i,ACTION);
    printf(":=");
    showAtomVal(i,pav->Val,ACTION);
    printf("\n");
  }
}

void WFShowTrans() {
  NUM i;
  printf("%d)\n",Now);
  for (i=0;i<NActions;i++)
    ShowActionVal(&Actions[i].Curr,i);
  for (i=0;i<NFluents;i++)
    ShowFluentVal(&Fluents[i].Curr,i);
}

int WFNewTrans()
{
  NUM i,v;
  tRULE *pr;  
  char ok=1,
       wasLoop; /* points out if the last change was a LoopDetection */

  #ifdef DEBUG
  printf("--- NEW TRANS Now=%d ---\n",Now);
  #endif

  ssInit(&CurRules,NRules);
  ssSequentialFill(CurRules,NRules);
  
  ssInit(&CurFluents,NFluents);

  #ifdef DEBUG
  printf("*** Solve dependences from previous values \n");
  #endif  

  for (i=0;i<NFluents;i++) {
    qBrowse(FluentDeps[i],SolvePrevDep,(void *)&i);    
    /* Insert fluents with layer 0 at top of CurFluents */
    if (!Fluents[i].NRemRules) 
      ssPush(CurFluents,i);
    else
      ssInsertLast(CurFluents,i);
  }
  #ifdef DEBUG
  PrintSS("CurRules",CurRules);
  PrintSS("CurFluents",CurFluents);
  printf("*** Perform actions\n");
  #endif  
  for (i=0;i<NActions;i++)
    qBrowse(ActionDeps[i],SolveActDep,(void *)&i);

  #ifdef DEBUG
  PrintSS("CurRules",CurRules);
  PrintSS("CurFluents",CurFluents);
  printf("*** Solving Fluents/Rules \n");
  #endif  

  wasLoop=0;
  for(;;) {
    
    if (!ssISEMPTY(CurRules)&&APPLICABLE(ssTOP(CurRules))) { 
      if ( !(ok=FireRule()) ) goto NewTrans_free_and_end;
      wasLoop=0;
      continue;
    }
    if (!ssISEMPTY(CurFluents)&&!Fluents[ssTOP(CurFluents)].NRemRules) { 
      if ( !(ok=Persistence()) ) goto NewTrans_free_and_end;
      wasLoop=0;
      continue;
    }
    if (   ssISEMPTY(CurRules)&&ssISEMPTY(CurFluents)
        || OptVal[optINFERENCE]==3 /* Fitting's semantics stops here */
        || wasLoop)
      break;
      
    LoopDetection();
    wasLoop=1;
  }
  
#ifdef DEBUG
  PrintSS("CurRules",CurRules);
  PrintSS("CurFluents",CurFluents);
  printf("--- END TRANS ---\n");
#endif  

  /* Check that no observation is required for remaining (undef.) fluents */
  for(i=ssTOP(CurFluents)+1;i!=0;i=CurFluents[i].next)
    if (Fluents[i-1].Obs!=NULL) 
      {
        ok=0; 
	if (verbose) {
	  printf("Inconsistence (in observation for fluent `");
	  showAtom(i-1,FLUENT);
	  printf("').\n"); 
	}
        goto NewTrans_free_and_end;
      }

NewTrans_free_and_end:
  free(CurRules);
  free(CurFluents);  

  return ok;
}

void FireDep(tFWDDEP *dep,char isPert)
{
  DECREMENT(Rules[dep->Rule]->F.NRemLits);
  if (isPert) 
    Rules[dep->Rule]->F.isPert=1; 
}

/************************************************************************/


#ifdef DEBUG
void ShowUnfoundedRule(tFWDDEP *dep, void *data,char Type)
{
  if(dep->Modifier==modPERT) printf("pert_");
  showAtom(*(NUM*)data,Type);
  if(dep->Sit!=0) printf("@%d",dep->Sit);
  printf("=");
  showAtomVal(*(NUM*)data,dep->Value,Type);
  printf(" is false: R%d unfounded\n",dep->Rule);    
}
#endif

void UpdateCurRules(NUM r) 
{
  /* if the rule becames unfounded ...*/
  if (Rules[r]->Unfounded) {
    /* we delete it from current rules */
    ssDelete(CurRules,r);
    /* and update the Remainig Rules counter for its consequent */
    if (!Rules[r]->FalseConseq)
      if (!--Fluents[Rules[r]->Consequent.Atom].NRemRules)
        /* if NRemRules becomes 0, move the fluent to the top */
        ssMoveTop(CurFluents,Rules[r]->Consequent.Atom);
    return;	
  }
  
  /* if the rule became applicable, it is moved to the top */
  if (APPLICABLE(r)) {
    #ifdef DEBUG
    printf("Rule R%d on top: prepared to be fired\n",r);
    #endif
    ssMoveTop(CurRules,r);
  }

}

int SolvePrevDep(QUEUE q,void *data)
{
  tFWDDEP *dep;
  tFLUENT *pf;
  tFLUENTVAL *Prev;

  pf=&Fluents[*(NUM *)data];
  dep=qFirst(q,tFWDDEP);

  /* if unfounded or not previous go out */
  if (Rules[dep->Rule]->Unfounded || dep->Sit==0) return 0;  
  
  switch (dep->Modifier) {
    case modNONE:
      if (pf->Prev==NULL)
	Rules[dep->Rule]->Unfounded=1;
      else {
	Prev=qFirst(pf->Prev,tFLUENTVAL);
	/* If prev. value was not N reached */
	if (! bvVal(Prev->N.Val,dep->Value))
	  Rules[dep->Rule]->Unfounded=1;
	else
	  if (Prev->isKnown && Prev->F.Val==dep->Value)
            Rules[dep->Rule]->F.NRemLits--;  
      }
      break;
    case modPERT:
      if (pf->Prev==NULL)
	Rules[dep->Rule]->Unfounded=1;
      else {
	Prev=qFirst(pf->Prev,tFLUENTVAL);
	switch (dep->Value) {
	  case 0:
	    if (!Prev->N.isPert) Rules[dep->Rule]->F.NRemLits--;
	    else 
	      if (Prev->F.isPert) Rules[dep->Rule]->Unfounded=1;  
	    break;
	  case 1:
	    if (!Prev->N.isPert) Rules[dep->Rule]->Unfounded=1;
	    else 
	      if (Prev->F.isPert) Rules[dep->Rule]->F.NRemLits--;  
	    break;
	}
      }
  }
  
  UpdateCurRules(dep->Rule);
  #ifdef DEBUG
  if (Rules[dep->Rule]->Unfounded) ShowUnfoundedRule(dep,data,FLUENT);
  #endif

  return 0;
}

int SolveActDep(QUEUE q,void *data)
{
  tFWDDEP *dep;
  tACTION *pa;
  tACTIONVAL *Prev;

  pa=&Actions[*(NUM *)data];
  dep=qFirst(q,tFWDDEP);

  /* if unfounded, go out */
  if (Rules[dep->Rule]->Unfounded) return 0;  

  switch (dep->Modifier) {
    case modNONE:
      if (dep->Sit==0) {
        Rules[dep->Rule]->F.isPert=1;
	/* check that the action has such a value */
	if (!(pa->Curr.isPert && pa->Curr.Val==dep->Value))
	  Rules[dep->Rule]->Unfounded=1;
      }
      else 
        if (pa->Prev==NULL)
	  Rules[dep->Rule]->Unfounded=1;
	else
	{
          Prev=qFirst(pa->Prev,tACTIONVAL);
	  /* check that the action had such value */
	  if (!(Prev->Sit==Now-1 && Prev->Val==dep->Value))
	    Rules[dep->Rule]->Unfounded=1;
	}
      break;
    case modPERT:
      if (dep->Sit==0) {
	if (dep->Value!=pa->Curr.isPert)
	  Rules[dep->Rule]->Unfounded=1;
        
	/* If action "pertinent", then rule condition becomes pertinent */
        if (pa->Curr.isPert)
          Rules[dep->Rule]->F.isPert=1;
	else
	/* if action not pertinent, decrement NRemPerts */  
	  if (!Rules[dep->Rule]->Unfounded) {
	    Rules[dep->Rule]->NRemPerts--;
	    if (!Rules[dep->Rule]->NRemPerts)
	      Rules[dep->Rule]->Unfounded=1;
	  }
      }
      else
        if (pa->Prev==NULL) {
	  if (dep->Value==1)
	    Rules[dep->Rule]->Unfounded=1;
	}
	else
	{
          Prev=qFirst(pa->Prev,tACTIONVAL);
	  /* check that the action had such value */
	  if (dep->Value!=(Prev->Sit==Now-1))
	    Rules[dep->Rule]->Unfounded=1;
	}
      break;
  }


  UpdateCurRules(dep->Rule);
  #ifdef DEBUG
  if (Rules[dep->Rule]->Unfounded) ShowUnfoundedRule(dep,data,ACTION);
  #endif

  return 0;
}


/************************************************************************
  We delete that rule whose condition becomes either founded or unfounded
*************************************************************************/

int FireRule()
{
  NUM r,v;
  tFLUENT *pf;
  
  r=ssTOP(CurRules);
  ssPop(CurRules);

#ifdef DEBUG
  printf("  firing rule R%d ==> ",r);
  if (Rules[r]->FalseConseq) printf("false\n");
  else {
    showAtom(Rules[r]->Consequent.Atom,FLUENT);
    printf(":=%d\n",Rules[r]->Consequent.Value);
  }
#endif
  
  if (Rules[r]->FalseConseq) {
    if (verbose)  {
      unsigned i=GetRulNum(r);
      tRulInfo *ri=(tRulInfo*)stEntry(tabRulInfo,i);
      printf("Inconsistence (rule %d, line %d, applied with `false' head).\n",
	i+1,ri->line); 
    }
    return 0;
  }  

  pf=&Fluents[Rules[r]->Consequent.Atom];
  if (pf->Curr.F.isPert && pf->Curr.isKnown && pf->Curr.F.Val!=Rules[r]->Consequent.Value)
  {
    if (verbose) {
      printf("Inconsistence (different pertinent values for fluent `");
      showAtom(Rules[r]->Consequent.Atom,FLUENT);
      printf("').\n"); 
    }
    return 0;
  }  
    
  pf->Curr.isKnown=1;
  pf->Curr.F.Val=Rules[r]->Consequent.Value;
  pf->Curr.F.isPert=pf->Curr.N.isPert=1;

  /* check observations */
  if (pf->Obs!=NULL && 
       (pf->Obs->isKnown && pf->Obs->F.Val!=pf->Curr.F.Val ||
        pf->Obs->F.isPert && !pf->Curr.F.isPert ||
	!pf->Obs->N.isPert && pf->Curr.N.isPert)
     ) {
    if (verbose) {
      printf("Inconsistence (in observation for fluent `");
      showAtom(Rules[r]->Consequent.Atom,FLUENT);
      printf("').\n"); 
    }
    return 0;
  }  

  /* Undo NVal marks  */
  for (v=0;v<pf->Curr.F.Val;v++)
    bvSet(pf->Curr.N.Val,v,0);
  bvSet(pf->Curr.N.Val,v++,1); /* ... excepting current value */
  for ( ; v<pf->NumVal; v++)
    bvSet(pf->Curr.N.Val,v,0);

  qBrowse(FluentDeps[Rules[r]->Consequent.Atom],
                SolveFluDep,&Rules[r]->Consequent.Atom);
  
  ssDelete(CurFluents,Rules[r]->Consequent.Atom);
  return 1;
}


int Persistence()
{
  NUM f;
  tFLUENT *pf;
  tFLUENTVAL *Prev;
  
  f=ssTOP(CurFluents);
  ssPop(CurFluents);

  pf=&Fluents[f];

  pf->Curr.N.isPert=pf->Curr.F.isPert=0;
  if (pf->Prev!=NULL) {
    Prev=qFirst(pf->Prev,tFLUENTVAL);
    pf->Curr.isKnown=Prev->isKnown;
    pf->Curr.F.Val=Prev->F.Val;
  }
  else
    pf->Curr.isKnown=0;

#ifdef DEBUG    
  printf("persistence ");
  showAtom(f,FLUENT);
  printf("=");
  if (pf->Curr.isKnown)
    showAtomVal(f,pf->Curr.F.Val,FLUENT);
  else
    printf("?");
  printf("\n");
#endif 

  /* check observations */
  if (pf->Obs!=NULL && 
       (pf->Obs->isKnown && pf->Obs->F.Val!=pf->Curr.F.Val ||
        pf->Obs->F.isPert && !pf->Curr.F.isPert ||
	!pf->Obs->N.isPert && pf->Curr.N.isPert) 
     ) 
  {
    if (verbose) {
      printf("Inconsistence (in observation for fluent `");
      showAtom(f,FLUENT);
      printf("').\n"); 
    }
    return 0;
  }  

  qBrowse(FluentDeps[f],SolveFluDep,&f);

  ssDelete(CurFluents,f);
  return 1;
}

/************************************************************************/

int SolveFluDep(QUEUE q,void *data)
{
  NUM r;
  tFWDDEP *dep;
  tFLUENT *pf;
  tFLUENTVAL *Prev;

  pf=&Fluents[*(NUM *)data];
  dep=qFirst(q,tFWDDEP);

  /* if unfounded or prev-dependence, go out */
  if (Rules[dep->Rule]->Unfounded || dep->Sit!=0) return 0;  

  /* In each case,
     if founded, fire the dependence (decrement F.NRemLits) 
     if unfounded, mark the whole rule as unfounded  */

  switch (dep->Modifier) {
    case modNONE:
      if (pf->Curr.isKnown)
        if (dep->Value==pf->Curr.F.Val) 
	  FireDep(dep,pf->Curr.F.isPert);
	else 
	  Rules[dep->Rule]->Unfounded=1;
      break;
    case modPERT:
      switch(dep->Value) {
	case 0:
	  if (!pf->Curr.N.isPert) FireDep(dep,0); 
	  else Rules[dep->Rule]->Unfounded=1;
	  break;
	case 1: 
	  if (pf->Curr.F.isPert) FireDep(dep,1);
	  else Rules[dep->Rule]->Unfounded=1;
      }
      break;
  }

  /* If the fluent persists, decrement the rule */
  if (!pf->Curr.N.isPert && !Rules[dep->Rule]->Unfounded) {
    Rules[dep->Rule]->NRemPerts--;
    if (!Rules[dep->Rule]->NRemPerts)
      Rules[dep->Rule]->Unfounded=1;
  }

  UpdateCurRules(dep->Rule);
  #ifdef DEBUG
  if (Rules[dep->Rule]->Unfounded) ShowUnfoundedRule(dep,data,FLUENT);
  #endif

  return 0;
}

/***** LOOP DETECTION ***************************************************/

int ResetDone(QUEUE q,void *d)
{
  qFirst(q,tFWDDEP)->Done=0;  
  return 0;
}

void NFireDep(tFWDDEP *dep,char isPert)
{
  /* if the dependence was not already done */
  if (!dep->Done) {
    DECREMENT(Rules[dep->Rule]->N.NRemLits);
    dep->Done=1;
    #ifdef DEBUG
    printf("      R%d remaining literals %d isPert %d\n",dep->Rule,Rules[dep->Rule]->N.NRemLits,Rules[dep->Rule]->N.isPert);
    #endif DEBUG
  }
  /* if the fluent became pertinent, mark the rule as pertinent */
  if (isPert) 
    Rules[dep->Rule]->N.isPert=1; 
}

int NSolveFluDep(QUEUE q,void *data)
{
  NUM r;
  tFWDDEP *dep;
  tFLUENT *pf;
  tFLUENTVAL *Prev;

  pf=&Fluents[*(NUM *)data];
  dep=qFirst(q,tFWDDEP);

  /* if prev-dependence or the rule was fired, go out */
  if (dep->Sit!=0 || !Rules[dep->Rule]->Unfounded) return 0;  
    
  switch (dep->Modifier) {
    case modNONE:
      /* if the value coincides */
      if (bvVal(pf->Curr.N.Val,dep->Value)) NFireDep(dep,pf->Curr.N.isPert);
      break;
    case modPERT:
      switch(dep->Value) {
	case 0:
	  /* all the nonpertinence atoms are true in loop detection */
	  NFireDep(dep,pf->Curr.N.isPert);
	  break;
	case 1: 
	  if (pf->Curr.N.isPert) NFireDep(dep,pf->Curr.N.isPert);
	  break;
      }
      break;
  }

  if (N_APPLICABLE(dep->Rule)) ssMoveTop(CurRules,dep->Rule);
  return 0;
}


void NFireRule()
{
  NUM r,a,v;
  tFLUENT *pf;
  
  r=ssTOP(CurRules);

  /* the rule will not be deleted */
  Rules[r]->Unfounded=0;
  ssMoveBottom(CurRules,r);
  
#ifdef DEBUG
  printf("  N-firing rule R%d ==> ",r);
  if (!Rules[r]->FalseConseq) {
    showAtom(Rules[r]->Consequent.Atom,FLUENT);
    printf(":=%d\n",Rules[r]->Consequent.Value);
  }
  else
    printf("false\n");
#endif

  if (Rules[r]->FalseConseq) return;

  a=Rules[r]->Consequent.Atom;
  v=Rules[r]->Consequent.Value;
  /* The fluent will go on undefined */
  ssMoveBottom(CurFluents,a);
  pf=&Fluents[a];
  bvSet(pf->Curr.N.Val,v,1);
  pf->Curr.N.isPert=1;
  pf->NRemRules++;
  
  qBrowse(FluentDeps[a],NSolveFluDep,(void*)&a);
}

void LoopDetection() 
{
  NUM i,j,r;
  char ok=1;
    
  #ifdef DEBUG
    PrintSS("CurRules",CurRules);
    PrintSS("CurFluents",CurFluents);
    printf("--------Positive loop detection\n");
  #endif

  /* Any fired rule (!Unfounded) will be put on the bottom of CurRules */
  /* Any "fired fluent" (N.isPert) will be put on the bottom of CurFluents */

  /* Set literal counters and nonpertinence for all rule conditions */
  for(i=ssTOP(CurRules)+1;i!=0;i=CurRules[i].next) {
    Rules[i-1]->N.NRemLits=Rules[i-1]->F.NRemLits;
    Rules[i-1]->N.isPert=(Rules[i-1]->F.isPert)?1:0;
    /* fired rules will move Unfounded to 0 */
    Rules[i-1]->Unfounded=1;
    /* unfounded rules will be on top of the stack, so that they can be deleted later */
  }

  for(i=ssTOP(CurFluents)+1;i!=0;i=CurFluents[i].next) {
    j=i-1;
    /* All the current fluents are nonpertinent... */
    Fluents[j].Curr.N.isPert=Fluents[j].Curr.F.isPert=0;

    /* Remaining rules is set to 0: each time we fire a rule it'll be incremented */
    Fluents[j].NRemRules=0;
    
    /* Reset the Done marks */
    qBrowse(FluentDeps[j],ResetDone,NULL);    
    
    /* Apply persistence */
    free(Fluents[j].Curr.N.Val);    
    if (Fluents[j].Prev!=NULL) {
      Fluents[j].Curr.isKnown=1;
      Fluents[j].Curr.F.Val=qFirst(Fluents[j].Prev,tFLUENTVAL)->F.Val;
      Fluents[j].Curr.N.Val=
        bvCopy(qFirst(Fluents[j].Prev,tFLUENTVAL)->N.Val,Fluents[j].NumVal);
    }
    else {
      Fluents[j].Curr.isKnown=0;
      Fluents[j].Curr.N.Val=bvNew(Fluents[j].NumVal);
    }
    
    #ifdef DEBUG
    printf("Fluent ");
    showAtom(j,FLUENT);
    printf(" n-persists\n");
    #endif
    qBrowse(FluentDeps[j],NSolveFluDep,(void *)&j);    
  }
  
  /* Fire rules */
  while (!ssISEMPTY(CurRules) && N_APPLICABLE(ssTOP(CurRules)) 
         && Rules[ssTOP(CurRules)]->Unfounded) 
  {	 
    NFireRule();
  }

  /* persistent fluents remain on top: they'll be deleted (NRemLits==0) */
  /* unfounded rules remain on top: they'll be deleted (Unfounded==1) */
  
  #ifdef DEBUG
    printf("--------End of positive loop detection\n");
  #endif

}

/************************************************************************/

int WFGetCurValue(struct tVarLstNode *pn,struct Elem *pe)
/* returns 0 if value unknown (or undefined because of bad argument)
   returns 1 otherwise
*/
{
  struct tFunctHeader *pf;
  NUM n;
  tFLUENTVAL *Prev;
  tACTIONVAL *PrevA;

  if (Fluents==NULL) {
    printf("%d: ",yyline); 
    printf("Cannot perform `query' before an `initially' clause.\n");
    exit(1);
  }

  n=GetAtomNum(pn->d.Funct);
  if (pal_code!=errNONE) return 0;
  switch (pn->Type) {
    case ACTION:
      switch(pn->d.Funct->Modifier) {
        case modNONE:
	  if (pn->d.Funct->Sit==0) {
	    if (!Actions[n].Curr.isPert) return 0;
	    pf=((struct tAtom *)stEntry(TabAct,n))->pfh;
	    memcpy(pe,&pf->Range[Actions[n].Curr.Val],sizeof(struct Elem));
	  }
	  else {
	    if (Actions[n].Prev==NULL) return 0;
	    PrevA=qFirst(Actions[n].Prev,tACTIONVAL);
	    if (PrevA->Sit!=Now-1) return 0;
	    pf=((struct tAtom *)stEntry(TabAct,n))->pfh;
	    memcpy(pe,&pf->Range[PrevA->Val],sizeof(struct Elem));
	  }
	  break;
        case modPERT:
	  pe->Type=SVALUE;
	  if (pn->d.Funct->Sit==0) {
	    if (Actions[n].Curr.isPert) 
	      pe->N.Sym=TRUE_NSym;
	    else
	      pe->N.Sym=FALSE_NSym;
	  }
	  else {
	    if (Actions[n].Prev==NULL)
	      pe->N.Sym=FALSE_NSym;
	    else {
	      PrevA=qFirst(Actions[n].Prev,tACTIONVAL);
	      if (PrevA->Sit==Now-1) pe->N.Sym=TRUE_NSym;
	      else pe->N.Sym=FALSE_NSym;
	    }
	  }
	  break;
      }
      break;
    case FLUENT:  
      switch(pn->d.Funct->Modifier) {
        case modNONE:
	  if (pn->d.Funct->Sit==0) {
	    if (!Fluents[n].Curr.isKnown) return 0;
	    pf=((struct tAtom *)stEntry(TabFlu,n))->pfh;
	    memcpy(pe,&pf->Range[Fluents[n].Curr.F.Val],sizeof(struct Elem));
	  }
	  else {
	    if (qIsEmpty(Fluents[n].Prev)) return 0;
            Prev=qFirst(Fluents[n].Prev,tFLUENTVAL);
	    if (!Prev->isKnown) return 0;
	    pf=((struct tAtom *)stEntry(TabFlu,n))->pfh;
	    memcpy(pe,&pf->Range[Prev->F.Val],sizeof(struct Elem));
	  }
	  break;
        case modPERT:
	  if (pn->d.Funct->Sit==0) {
	    if (!Fluents[n].Curr.F.isPert && Fluents[n].Curr.N.isPert) return 0;	   
	    pe->Type=SVALUE;
	    if (Fluents[n].Curr.F.isPert) 
	      pe->N.Sym=TRUE_NSym;
	    else
	      pe->N.Sym=FALSE_NSym;
	  }
	  else {
	    if (qIsEmpty(Fluents[n].Prev)) return 0;
            Prev=qFirst(Fluents[n].Prev,tFLUENTVAL);
	    if (!Prev->F.isPert && Prev->N.isPert ) return 0;
	    pe->Type=SVALUE;
	    if (Prev->Sit==Now-1 && Prev->F.isPert) pe->N.Sym=TRUE_NSym;
	    else pe->N.Sym=FALSE_NSym;	    
	  }
	  break;
      }
  }
  return 1;
}

int WFGetValue(NUM n,char Type,NUM *v,char *isPert)
/* returns 0 if undefined because of bad argument or Fluents is NULL
   returns 1 otherwise
*/
{
  if (Fluents==NULL) return 0;

  switch(Type) {
    case ACTION:
      *isPert=Actions[n].Curr.isPert;
      *v=Actions[n].Curr.Val;
      return 1;

    case FLUENT:  
      if (!Fluents[n].Curr.isKnown) return 0;
      if (Fluents[n].Curr.N.isPert!=Fluents[n].Curr.F.isPert) return 0;
      *isPert=Fluents[n].Curr.F.isPert;
      *v=Fluents[n].Curr.F.Val;
      return 1;
  }
}

int FluInsert1st(QUEUE q,void *p)
{
  if (qFirst(q,tFLUENTVAL)->Sit<=IniSit) return 1;
  qInsertFirst((QUEUE*)p,qFirst(q,tFLUENTVAL),sizeof(tFLUENTVAL));
  return 0;
}

int ActInsert1st(QUEUE q,void *p)
{
  if (qFirst(q,tACTIONVAL)->Sit<=IniSit) return 1;
  qInsertFirst((QUEUE*)p,qFirst(q,tACTIONVAL),sizeof(tACTIONVAL));
  return 0;
}

void WFShowSolution()
{
  NUM i;
  SIT s;
  char first=1;
  QUEUE *FluData=NULL,*ActData=NULL;
  void *p;
  
  NSol++;
  if(ExprSit>1)
    printf("\nSolution %d:\n",NSol);
  for(i=0;i<NVars;i++)
    if (VarVec[i].Type==VAR)
    {
      if (first) first=0; else printf(",");
      printf("%s",stEntry(SymTables[VAR],VarVec[i].d.Sym));
      printf("=");
      showElem(&tup[i]);
    }
  if (!first) printf("\n");
  
  if (ExprSit<=1) return; /* no temporal query */

  /* Reverse data lists */
  ActData=(QUEUE*)calloc(NActions,sizeof(QUEUE));
  FluData=(QUEUE*)calloc(NFluents,sizeof(QUEUE));


  
  for(i=0;i<NFluents;i++) {
    if(qIsNotEmpty(Fluents[i].Prev))
      qBrowse(Fluents[i].Prev,FluInsert1st,&FluData[i]);
    if(Fluents[i].Curr.N.isPert)
      qInsertLast(&FluData[i],&Fluents[i].Curr,sizeof(tFLUENTVAL));    
  }

  for(i=0;i<NActions;i++) {
    if(qIsNotEmpty(Actions[i].Prev))
      qBrowse(Actions[i].Prev,ActInsert1st,&ActData[i]);      
    if(Actions[i].Curr.isPert)
      qInsertLast(&ActData[i],&Actions[i].Curr,sizeof(tACTIONVAL));
  }

  for (s=IniSit+1;s<IniSit+ExprSit;s++)
  {
    printf("%d)\n",s);
    for(i=0;i<NActions;i++)
      if (qIsNotEmpty(ActData[i]) && qFirst(ActData[i],tACTIONVAL)->Sit==s) {
        ShowActionVal(qFirst(ActData[i],tACTIONVAL),i);
        qDeleteFirst(&ActData[i]);
      } 
    for(i=0;i<NFluents;i++)
      if (qIsNotEmpty(FluData[i]) && qFirst(FluData[i],tFLUENTVAL)->Sit==s) {
        ShowFluentVal(qFirst(FluData[i],tFLUENTVAL),i);
        qDeleteFirst(&FluData[i]);
      }
  }
  free(FluData);
  free(ActData);  
}
